#
# Copyright (c) 2015-2022, Sine Nomine Associates ("SNA")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND SNA DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL SNA BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from fnmatch import fnmatch
import json
import time

from fabs.dumpjob import DumpJob

import fabs.server
import fabs.report as report
import fabs.util as util
import fabs.err as err
import fabs.db as db
import fabs.config as config
import fabs.bstore as bstore
import fabs.log
log = fabs.log.getLogger(__name__)

class BackupRun:
    """
    A BackupRun represents a fabs backup run, a single run through the cell,
    backing up a single volume or a set of volumes at a particular time.

    Don't instantiate a BackupRun directly. You can create a new one with
    BackupRun.create(), or pull existing ones from the db with
    BackupRun.find().
    """

    @classmethod
    def create(cls, cell, note, volume, injected_blob=None):
        """
        Create a new BackupRun.

        Args:
            cell (str): The cell to backup.
            note (str): Human-readable note (e.g. "Daily backup for xxx").
            volume (str or None): The name of the volume to backup, for a
                single-volume backup. Give `None` to backup all volumes
                according to the fabs config.
            injected_blob (`fabs.bstore.DumpBlob`, optional): If provided, this
                brun will use the given dump blob instead of dumping the volume
                from the fileserver.

        Returns:
            `BackupRun`: The created backup run.
        """
        iblob_storid = None
        iblob_spath = None
        iblob_cksum = None

        if injected_blob:
            if not volume:
                raise err.InternalError("injected_blob without volume")

            iblob_storid = injected_blob.storid()
            iblob_spath = injected_blob.rel_path()
            iblob_cksum = injected_blob.checksum()
            injected_blob.no_delete()

        brid = db.brun.create(cell=cell, note=note, volume=volume,
                              injected_blob_storid=iblob_storid,
                              injected_blob_spath=iblob_spath,
                              injected_blob_checksum=iblob_cksum)
        return cls.find(id=brid)[0]

    @classmethod
    def _binfo_status(cls, binfo):
        ret = ''

        data = {}
        data.update(binfo)

        if data['state_last'] is not None:
            data['state'] += "/%s" % data['state_last']

        if data['volume'] is None:
            data['volume'] = '*'

        for field in ('start', 'end'):
            if int(data[field]) > 0:
                data[field] = util.time_unix2str(int(data[field]))

        data['jobs_total'] = len(data['jobs'])

        ret += """Backup run %(id)d [%(state)s]: %(note)s
  status.%(dv)d: %(state_descr)s
  cell: %(cell)s, volume: %(volume)s, errors: %(errors)d
  start: %(start)s
  end:   %(end)s
  jobs: %(jobs_total)d total
""" % data

        collapse_states = ('NEW', 'CLONE_PENDING', 'CLONE_DONE',
                           'DUMP_PENDING', 'DUMP_SKIPPED', 'DUMP_DONE',
                           'DONE')

        state_summary = {}
        expand_jobs = {}

        jobs = binfo['jobs']
        for job in jobs:
            if job['state'] in collapse_states:
                state_summary[job['state']] = state_summary.get(job['state'], 0) + 1
            else:
                key = job['server_addr']
                expand_jobs.setdefault(key, [])
                expand_jobs[key].append(job)

        for state in collapse_states:
            numjobs = state_summary.get(state, 0)
            if numjobs > 0:
                ret += "  [%d job(s) in state %s]\n" % (numjobs, state)

        ret += "\n"

        for server_addr, jobs in expand_jobs.items():

            ret += "  Server %s, %d job(s):\n" % (server_addr, len(jobs))

            for job in jobs:
                data = {}
                data.update(job)
                if data['state_last'] is not None:
                    data['state'] += "/%s" % data['state_last']

                for field in ('ctime', 'mtime'):
                    if int(data[field]) > 0:
                        data[field] = time.ctime(int(data[field]))

                ret += """    vol %(name)s job %(vl_id)d.%(dv)d [%(state)s]: %(state_descr)s (%(state_source)s)
      ctime: %(ctime)s, rwid: %(rwid)d, part: vicep%(partition)s
      mtime: %(mtime)s, timeout: %(timeout)d, errors: %(errors)d

""" % data

        return ret

    @classmethod
    def status(cls, brid=None, failed=False):
        """
        Get the status of the specified backup runs.

        Args:
            brid (int, optional): If given, return the status of the brun with
                the given brid. Otherwise, return the status of failed or
                active bruns (depending on the other arguments given).
            failed (bool, optional): If True, return the status of all failed
                bruns. Otherwise, and if `brid` is not given, return the status
                of active bruns. Defaults to False.

        Returns:
            dict: A dictionary containing status info for the requested backup
            runs.
        """

        infos = []

        with db.connect(lock=False):
            if brid is not None:
                brun_list = cls.find(id=brid)
            elif failed:
                brun_list = cls.find(state='FAILED')
            else:
                brun_list = cls.find(state_not=['DONE', 'FAILED'])

            for brun in brun_list:
                bruninfo = {}
                # pylint: disable=not-an-iterable
                for col in db.model.backup_runs.columns:
                    attr = col.name
                    bruninfo[attr] = getattr(brun, attr)

                jobs = [jobinfo for jobinfo in DumpJob.describe(brun)]
                bruninfo['jobs'] = jobs

                infos.append(bruninfo)

        return {'status': infos}

    @classmethod
    def status_str(cls, info):
        """
        Translate the info from `BackupRun.status` into a human-readable
        string.

        Args:
            info (dict): A status dict returned from `BackupRun.status`.

        Returns:
            str: A human-readable formatted status.
        """

        info_list = info['status']
        ret = ''
        ret += "Found %d backup run(s)\n" % len(info_list)
        for binfo in info_list:
            ret += cls._binfo_status(binfo)
        return ret

    @classmethod
    def find(cls, **kwargs):
        """
        Find an existing BackupRun in the database.

        Args:
            **kargs: Search criteria. Any BackupRun database column can be
            specified; common criteria are, for example:

            * id=5

            * state='DONE'

            * state_not=['DONE','FAILED']

            See `fabs.db._BackupRunDb.find` for specifics on criteria that can
            be given.

        Returns:
            list of BackupRun: A list of matching bruns. Note that a list is
            returned even if you're just looking for a single brun by brid.
            (You get a list of length 1 in that case).

            If no bruns are found, and you didn't search for a specific id, an
            empty list is returned.

        Raises:
            `fabs.err.BadBridError`: If you searched for a specific id, but
            that id doesn't exist.
        """
        rows = db.brun.find(**kwargs)

        # If we were given an id, that backuprun should really exist.
        if 'id' in kwargs and not rows:
            raise err.BadBridError("Backup run id %d does not seem to exist" % kwargs['id'])

        return [BackupRun(row) for row in rows]

    @classmethod
    def find_update(cls, cell, from_state, to_state, descr):
        """
        Transition all bruns from one state to another state, and return all
        bruns in the new state.

        Args:
            cell (str):       Only update bruns for this cell.
            from_state (str): Search for bruns in this state.
            to_state (str):   Transition matching bruns to this state.
            descr (str):      For bruns in `from_state`, set the status description
                to this when transitioning to `to_state`.

        Returns:
            list of BackupRun: All bruns in `to_state`. This may include bruns
            that were already in `to_state`; we didn't necessarily transition
            them to `to_state` in this function.
        """

        bruns = cls.find(cell=cell, state=from_state)

        if to_state is None:
            return bruns

        with db.connect():
            for brun in bruns:
                if descr is None:
                    brun.update(state=to_state)
                else:
                    brun.update(state=to_state, state_descr=descr)

        bruns = cls.find(cell=cell, state=to_state)
        return bruns

    @classmethod
    def _filter_dups(cls, bruns):
        # Check if we have duplicates, so we don't run the same backup run more
        # than once in parallel

        # Keep track of a dict mapping 'volume' to the brun for that volume
        # (remember, a volume of None means the brun is for all volumes)
        vol_map = {}
        for brun in bruns:
            conflict_brun = vol_map.get(brun.volume, None)

            if conflict_brun is not None:
                brun.update(state='FAILED',
                            state_descr="Conflict with brid %d in state %s" % (
                                         conflict_brun.id, conflict_brun.state))
            else:
                vol_map[brun.volume] = brun

        # When we're done, all of the bruns in the vol_map should be unique wrt
        # volume
        return list(vol_map.values())

    @classmethod
    def _process(cls, cell, from_state, to_state, start_state, func, descr):
        """
        Process backup runs in a given state.

        For each brun in the given cell and `from_state`, transition them to
        `to_state` and set the status description to `descr`.

        Then for each brun in `to_state`, spawn a child process, transition the
        brun to `start_state`, and run `func`.

        Args:
            cell (str):        Match bruns in this cell.
            from_state (str):  Match bruns in this state.
            to_state (str):    Transition matching bruns to this state.
            start_state (str): Transition bruns to this state in child process.
            func (function):   Function to run in the child process.
            descr (str):       For bruns in `from_state`, set the status description
                to this when transitioning to `to_state`.
        """

        bruns = cls.find_update(cell, from_state, to_state, descr)
        log.d("Found %d bruns in state %s -> %s" % (len(bruns), from_state, to_state))

        bruns = cls._filter_dups(bruns)
        if not bruns:
            # no matching bruns; nothing to do
            return

        procs = []

        with db.connect_bulk() as tx:
            for brun in bruns:
                if start_state is not None:
                    brun.update(state=start_state)
                    tx.entry_done()

                if func:
                    procs.append(util.FabsProcess(target=brun.process_child, args=(func,)))

        for proc in procs:
            proc.start()

    @classmethod
    def process_jobs(cls):
        """
        Process work that needs to be done for backup runs.
        """

        cell = config.get('afs/cell')

        cls._process(cell, 'NEW', 'INIT_START', 'INIT_WORK', cls.do_init,
                     "Starting backup run initialization")

        # Scan the vldb for fileserver addresses, and store the addresses
        cls._process(cell, 'INIT_DONE', 'VLADDR_START', 'VLADDR_WORK', cls.do_vladdr,
                     "Starting to scan vldb server addresses")

        # Scan the vldb for volume entries, and store the ones we care about
        cls._process(cell, 'VLADDR_DONE', 'VLENTRY_START', 'VLENTRY_WORK',
                     cls.do_vlentry, "Starting to scan vldb entries")

        # Create our jobs to clone volumes. (Not scheduled, just created in
        # the db.)
        cls._process(cell, 'VLENTRY_DONE', 'BRUN_CLCREATE_START', 'BRUN_CLCREATE_WORK',
                     cls.create_clones, "Starting volume clones")

        # Schedule any pending clone jobs. This runs repeatedly, scheduling
        # more jobs as we are able, respecting our various parallel quotas.
        cls._process(cell, 'BRUN_CLCREATE_DONE', 'BRUN_CLSCHED_START', 'BRUN_CLSCHED_WORK',
                     cls.sched_clones, "About to schedule volume clones")

        # Create our jobs to dump volumes. (Not scheduled, just created in the
        # db.)
        cls._process(cell, 'BRUN_CLSCHED_DONE', 'BRUN_DUMPCREATE_START', 'BRUN_DUMPCREATE_WORK',
                     cls.create_dumps, "Starting volume dumps")

        # Schedule any pending dump jobs. This runs repeatedly, scheduling more
        # jobs as we are able, respecting our various parallel quotas.
        cls._process(cell, 'BRUN_DUMPCREATE_DONE', 'BRUN_DUMPSCHED_START', 'BRUN_DUMPSCHED_WORK',
                     cls.sched_dumps, "About to schedule volume dumps")

        # Send a report out about this finished backup run, including what
        # backups succeeded, failed, etc.
        cls._process(cell, 'BRUN_DUMPSCHED_DONE', 'BRUN_REPORT_START', 'BRUN_REPORT_WORK',
                     cls.do_report, "About to report on finished backup run")

        cls._process(cell, 'BRUN_REPORT_DONE', 'BRUN_CLEANUP_START', 'BRUN_CLEANUP_WORK',
                     cls.do_cleanup, "Cleaning up finished backup run")

        # BRUN_REPORT_DONE is the last stage; we don't have anything more to
        # do. So move the brun into the DONE stage to indicate there's nothing
        # more. Don't use _process for this, since we don't want to find all
        # DONE bruns, and we don't need to spawn a new process just for this
        # final transition.
        cls._finish_bruns(cell, 'BRUN_CLEANUP_DONE')

        # Retry stages that failed with an error
        cls._process(cell, 'ERROR', 'BRUN_RECOVER_START', 'BRUN_RECOVER_WORK',
                     cls.do_recover, None)

        # For recent runs, check if they have any dangling vl_entry rows or
        # other cruft.
        for brun in cls.cleanable_bruns(recent=True):
            brun.clean_cruft()

    @classmethod
    def _finish_bruns(cls, cell, state):
        for brun in cls.find(cell=cell, state=state):
            brun.update(state='DONE', state_descr="Done")
            log.info('brun_done', "Backup run %d finished" % brun.id)

    @classmethod
    def cleanable_bruns(cls, recent=False, all_=False):
        """
        Find backup runs that are eligible for cleaning up.

        You must specify either `recent` or `all_`.

        Args:
            recent (bool, optional): If True, only look for recent bruns
                (started in the last 12 hours). If there are any active bruns,
                don't find any bruns (since fabs is busy running those bruns).
                Defaults to False.
            all_ (bool, optional): If True, look for all bruns that aren't
                actively running. Defaults to False.

        Returns:
            list of `BackupRun`: A list of matching bruns.
        """

        if recent:
            active_bruns = cls.find(state_not=['DONE','FAILED'])
            if active_bruns:
                brun = active_bruns[0]
                log.d("Not cleaning bruns, brun %d is active" % (brun.id))
                return []

            now = int(time.time())
            interval = 60*60*12 # twelve hours
            start_time = now-interval

            return cls.find(start_after=start_time)

        elif all_:
            return cls.find(state=['DONE','FAILED'])

        raise ValueError("We're not cleaning all backup_runs, but no criteria given?")

    def __init__(self, row):
        """
        Initialize a BackupRun object.

        Don't run this directly; use `BackupRun.create`, or `BackupRun.find`,
        or something else.

        Args:
            row (`sqlalchemy.engine.Row`): The row from the database that
                represents the BackupRun to instantiate.
        """
        for col in row.keys():
            setattr(self, col, row[col])

    def describe_dict(self):
        """
        Returns:
            dict: A dictionary of info describing this brun.
        """
        brinfo = {}
        # pylint: disable=not-an-iterable
        for col in db.model.backup_runs.columns:
            attr = col.name
            brinfo[attr] = getattr(self, attr)
        return brinfo

    def describe(self, from_dict=None):
        """
        Args:
            from_dict (dict, optional): A dict from `BackupRun.describe_dict`,
                describing this brun. If not given, we'll just call
                `BackupRun.describe_dict` to get one.

        Returns:
            str: A human-readable formatted string describing this brun.
        """
        if from_dict is None:
            from_dict = self.describe_dict()
        data = {}
        data.update(from_dict)
        for field in ('start', 'end'):
            if int(data[field]) > 0:
                data[field] = util.time_unix2str(int(data[field]))
        return """Backup run %(id)d [%(state)s]: %(note)s
  status.%(dv)d: %(state_descr)s
  cell: %(cell)s, volume: %(volume)s, errors: %(errors)d
  start: %(start)s
  end:   %(end)s""" % data

    def delete(self):
        """
        Delete this brun from the database.

        This will delete all associated VL data for the backup run, but will
        not delete any existing dumps. If there are any dumps associated with
        this brun, this will fail due to database constraints.
        """
        db.brun.delete(self.id)

    def is_empty(self):
        """
        Returns:
            bool: True if this backup run is empty; that is, it has no volume
            dumps associated with it (and the caller can delete it). False
            otherwise.
        """
        for _ in db.voldump.finditer(brid=self.id):
            return False
        return True

    def update(self, **kwargs):
        """
        Update data for this brun in the db.

        Args:
            **kwargs: Columns to update in the database, and the values to
                update them to. See `fabs.db._BackupRunDb.update` for
                specifics.
        """
        db.brun.update(self, **kwargs)

    def error(self):
        """
        Flag this brun as encountering an error.

        Typically this function will be called when an exception is caught
        while trying to do something for this brun, or in general an error is
        encountered.

        This function will put this brun in an error state, unless it's already
        trying to recover from an error, in which case it will be put
        immediately into FAILED state.
        """

        if self.state in ('ERROR', 'BRUN_RECOVER_START', 'BRUN_RECOVER_WORK'):
            log.error('dump_err_meta', ("Backup run %d encountered an error " +
                      "while trying to recover from a previous error. It " +
                      "will not be retried.") % self.id)
            self.update(state='FAILED')
        else:
            db.brun.error(self)

    def kill(self, note):
        """
        Kill this brun.

        This puts the brun immediately into FAILED state.

        Args:
            note (str): A human-readable note to say why this brun was killed.
        """
        db.brun.kill(self, note)

    def retry(self):
        """
        Retry this failed/error'd brun.

        This tries to get this brun to run again, after encountering errors.
        This resets the error counter, and puts the brun in BRUN_RECOVER_START
        state, where it can go through the error recovery process.
        """

        if self.state not in ('ERROR', 'FAILED'):
            raise err.RetryStateError(("Backup run %d is not in an error " +
                                       "state (it's in %s)") % (
                                       self.id, self.state))
        self.update(state='BRUN_RECOVER_START', errors=0)

    def process_child(self, func):
        try:
            func(self)
        except Exception:
            log.exception('brun_child_err', "Error when processing backup run %d" % self.id)
            self.error()

    def do_recover(self):
        """
        Go through the error recovery process for this brun.

        This looks at what the last state was for this brun, and resets the
        brun back to the appropriate state so the relevant work can be retried.
        """

        ret_table = dict(
            VLADDR_WORK='VLADDR_START',
            VLENTRY_WORK='VLENTRY_START',
            BRUN_CLCREATE_WORK='BRUN_CLCREATE_START',
            BRUN_CLSCHED_WORK='BRUN_CLSCHED_START',
            BRUN_DUMPCREATE_WORK='BRUN_DUMPCREATE_START',
            BRUN_DUMPSCHED_WORK='BRUN_DUMPSCHED_START',
            BRUN_REPORT_WORK='BRUN_REPORT_START',
            BRUN_CLEANUP_WORK='BRUN_CLEANUP_START',

            # These are states where we can stay in the same state to recover.
            # That is, e.g. VLADDR_START just gets transitioned back to
            # VLADDR_START; we don't need to change to a new state.
            NEW=None,
            VLADDR_START=None,
            VLADDR_DONE=None,
            VLENTRY_START=None,
            VLENTRY_DONE=None,
            BRUN_CLCREATE_START=None,
            BRUN_CLCREATE_DONE=None,
            BRUN_CLSCHED_START=None,
            BRUN_CLSCHED_DONE=None,
            BRUN_DUMPCREATE_START=None,
            BRUN_DUMPCREATE_DONE=None,
            BRUN_DUMPSCHED_START=None,
            BRUN_DUMPSCHED_DONE=None,
            BRUN_REPORT_START=None,
            BRUN_REPORT_DONE=None,
            BRUN_CLEANUP_START=None,
            BRUN_CLEANUP_DONE=None,
        )

        failed = False

        if self.errors >= config.get('dump/error_limit'):
            log.error('brun_err_limit', ("Backup run %d has failed %d " +
                      "times, which exceeds our retry limit") % (
                      self.id, self.errors))
            failed = True

        elif self.state_last not in ret_table:
            log.error('brun_err_unknown', ("Backup run %d encountered " +
                      "an error in state %s, which we do not know how to " +
                      "retry") % (self.id, self.state_last))
            failed = True

        if failed:
            self.update(state='FAILED')
        else:
            retry_state = ret_table[self.state_last]
            if retry_state is None:
                retry_state = self.state_last

            log.warn('brun_retry', "Retrying backup run %d %s -> %s" % (
                     self.id, self.state_last, retry_state))
            self.update(state=retry_state, state_last=None)

            fabs.server.made_progress()

    def injected_blob(self):
        """
        Returns:
            `fabs.bstore.DumpBlob` or None: The injected dump blob for this
            brun, or None if there was none.
        """
        return bstore.blob_from_db(self.injected_blob_storid,
                                   self.injected_blob_spath,
                                   self.injected_blob_checksum)

    def do_cleanup(self):
        dblob = self.injected_blob()
        if dblob:
            # Get rid of our injected dump blob if it's still around
            dblob.delete()

        self.clean_cruft(active=True)

        self.update(state='BRUN_CLEANUP_DONE',
                    state_descr="Done cleaning up backup run")
        fabs.server.made_progress()

    def do_report(self):
        report.report_success(self)

        self.update(state='BRUN_REPORT_DONE', state_descr="Done sending report(s)")
        fabs.server.made_progress()

    def do_init(self):
        try:
            bstore.periodic_cleaning()
        except Exception:
            log.exception('bstore_cleanup_err', ("Error while performing " +
                          "periodic cleaning of blob storage dirs"))

        log.info('brun_start', "Starting backup run %d, backing up volumes: %s" % (
                 self.id, '*' if self.volume is None else self.volume))

        self.update(state='INIT_DONE', state_descr="Done initializing backup run")

        fabs.server.made_progress()

    def do_vladdr(self):

        self.update(state_descr="Scanning VLDB server addresses")

        vos = util.vos_unauth(cell=self.cell)
        res = vos.listaddrs(printuuid=True, noresolv=True).run()

        log.d("Adding %d vldb servers for brun %d" % (len(res.servers), self.id))

        db.vladdr.create_many(self, res.servers)

        self.update(state='VLADDR_DONE', state_descr="Done scanning VLDB server addresses")

        # Tell the server that we have made progress, so it will immediately
        # check for more stuff to do.
        fabs.server.made_progress()

    def _include_volume(self, vlentry, rw_fsid, inc_vol, exc_vol, inc_fs, exc_fs):
        """
        Determine whether or not to include the given volume in this backup
        run.

        This checks the vlentry for a volume, and sees if we should include it
        in the volumes to backup during this backup run, based on the
        per-volume and per-fileserver lists given.

        Args:
            vlentry (`fabs.cmd.VLDBEntry`): The VL entry to check.
            rw_fsid (int): ID number of the fileserver for the vol's RW site.
            inc_vol (list of str): fnmatch patterns of volume names to include.
            exc_vol (list of str): fnmatch patterns of volume names to exclude.
            inc_fs (list of int): Fileserver IDs to include.
            exc_fs (list of int): Fileserver IDs to exclude.

        Returns:
            bool: True if the volume should be backed up, False if the volume
            should be ignored.
        """

        matched = False

        if rw_fsid in inc_fs:
            log.d("volume %s matched inc_fs %r" % (vlentry.name, inc_fs))
            matched = True

        if not matched:
            for pat in inc_vol:
                if fnmatch(vlentry.name, pat):
                    log.d("volume %s matched vol pattern %s" % (vlentry.name, pat))
                    matched = True
                    break

        if not matched:
            log.d("Ignoring volume %s (not in include pattern)" % vlentry.name)
            return False

        if rw_fsid in exc_fs:
            log.d("Ignoring volume %s because it matched exc_fs %r" % (vlentry.name, exc_fs))
            return False

        for pat in exc_vol:
            if fnmatch(vlentry.name, pat):
                log.d("Ignoring volume %s (in exclude pattern %s)" % (vlentry.name, pat))
                return False

        log.d("brun %s including volume %s" % (self.id, vlentry.name))
        return True

    def _get_fsid(self, addrmap, server):
        fsid = addrmap.get(server, None)
        if fsid is None:
            log.warn('vlentry_noaddr',
                     "Server %s is unknown, creating a non-uuid entry for it" %
                     server)
            db.vladdr.create_nonuuid(self, addr=server)

            # Update our addrmap with whatever's in the db now. (Make sure to
            # alter addrmap, and not just assign to it, so our caller sees our
            # changes.)
            addrmap.clear()
            addrmap.update(db.vladdr.get_addrmap(self))

            fsid = addrmap.get(server, None)
            if fsid is None:
                raise err.VLFsIdError("Failed to create non-uuid entry for %s" %
                                      server)
        return fsid

    # "cludes" as in, 'cludes, as in includes/excludes.
    def _get_cludes(self):
        self.update(state_descr="Processing includes/excludes")

        cludes = {
            ('include', 'volumes'): [],
            ('exclude', 'volumes'): [],
            ('include', 'fileservers'): [],
            ('exclude', 'fileservers'): [],
        }

        # First, read cludes from e.g. config.get('dump/include/volumes')
        for (clude, clude_type) in cludes.keys():
            cludes[(clude, clude_type)].extend(config.get('dump/%s/%s' % (clude, clude_type)))

        # Then read in any cludes from our configured command, if any
        cmd = config.get('dump/filter_cmd/json')
        if cmd:
            raw_data = util.run_config_cmd(cmd)

            try:
                data = json.loads(raw_data)

                for (clude, clude_type) in cludes.keys():
                    if clude in data and clude_type in data[clude]:
                        cludes[(clude, clude_type)].extend(data[clude][clude_type])

            except Exception as exc:
                raise err.ConfigCmdError("Error processing data from filter command '%s': %s" % (
                                         cmd, type(exc).__name__ + str(exc)))

        log.info('filter', ("Backup run %d including volumes [%s], " +
                            "excluding volumes [%s], including " +
                            "fileservers [%s], excluding fileservers " +
                            "[%s]") % (
                            self.id,
                            ', '.join(cludes[('include', 'volumes')]),
                            ', '.join(cludes[('exclude', 'volumes')]),
                            ', '.join(cludes[('include', 'fileservers')]),
                            ', '.join(cludes[('exclude', 'fileservers')])))

        # Translate fileserver names into vl_fileservers ids
        uuidmap = db.vladdr.get_uuidmap(self)
        for (clude, clude_type) in [('include', 'fileservers'), ('exclude', 'fileservers')]:
            cludes[(clude, clude_type)] = self._fs2id(uuidmap, cludes[(clude, clude_type)])

        return (cludes[('include', 'volumes')],
                cludes[('exclude', 'volumes')],
                cludes[('include', 'fileservers')],
                cludes[('exclude', 'fileservers')],)

    def _fs2id(self, uuidmap, fslist):
        """
        Translate a list of fileserver names/addresses into their corresponding
        ids in the vl_fileservers table.

        We do this by running 'vos listaddrs' for each address to get the
        corresponding uuid, and then lookup the uuid in the given uuidmap.
        """
        vos = util.vos_unauth(cell=self.cell)

        ids = []

        for fs in fslist:
            log.d("Running listaddrs for fileserver %s" % fs)
            res = vos.listaddrs(host=fs, printuuid=True, noresolv=True).run()
            log.d("res: %r" % res)
            uuid = res.servers[0].uuid
            fs_id = uuidmap[uuid]
            ids.append(fs_id)
            log.d("Looked up fileserver %s as uuid %s fs_id %d" % (fs, uuid, fs_id))

        return ids

    def do_vlentry(self):

        (inc_vol, exc_vol, inc_fs, exc_fs) = self._get_cludes()

        self.update(state_descr="Scanning VLDB volume entries")

        vos = util.vos_unauth(cell=self.cell)
        if self.volume:
            cursor = vos.listvldb(name=self.volume, noresolv=True)
        else:
            cursor = vos.listvldb(noresolv=True)

        # Ideally we should probably have streaming results from this, but eh,
        # not now.
        res = cursor.run()

        log.d("Processing %d vldb entries" % len(res.entries))

        if not self.volume:
            log.d("Checking volumes against patterns include: %r exclude: %r" % (
                  config.get('dump/include/volumes'), config.get('dump/exclude/volumes')))

        addrmap = db.vladdr.get_addrmap(self)

        with db.connect_bulk() as tx:
            for entry in res.entries:

                rw_fsid = None
                for site in entry.sites:
                    if site.vtype == 'RW':
                        rw_fsid = self._get_fsid(addrmap, site.server)
                        break
                if rw_fsid is None:
                    log.warn('vlentry_norw', ("Cannot find an RW site for " +
                                              "volume %s; skipping it") % entry.name)
                    continue

                if not self.volume and not self._include_volume(entry, rw_fsid,
                                                                inc_vol,
                                                                exc_vol,
                                                                inc_fs,
                                                                exc_fs):
                    continue

                vl_id = db.vlentry.create(self, entry)
                vlsites = []

                for site in entry.sites:
                    fsid = self._get_fsid(addrmap, site.server)
                    part = site.partition.lower()
                    if part.startswith('/vicep'):
                        part = part[6:]

                    siteinfo = dict(
                        vl_id=vl_id,
                        type=site.vtype,
                        server_id=fsid,
                        partition=part
                    )

                    vlsites.append(siteinfo)
                db.vlentry.create_sites(entry, vlsites)

                tx.entry_done()

        self.update(state='VLENTRY_DONE', state_descr="Done scanning VLDB volume entries")

        fabs.server.made_progress()

    def create_clones(self):
        self.update(state='BRUN_CLCREATE_WORK', state_descr="Creating clone jobs")

        DumpJob.clear(self)

        with db.connect_bulk() as tx:
            for row in db.vlentry.finditer(self):
                rwsite = None
                for site in row['sites']:
                    if site['type'] == 'RW':
                        rwsite = site
                        break

                if rwsite is None:
                    log.warn('no_rw', "Volume %s has no RW site; not backing it up" %
                                      row['name'])
                    continue

                iblob_storid = None
                iblob_spath = None
                iblob_cksum = None
                if self.injected_blob_spath:
                    # We have an injected dump blob. We had better be processing
                    # the volume that this dump blob is for, otherwise something is
                    # screwy.
                    if row['name'] != self.volume:
                        raise err.InternalError("vl_entry name (%s) != brun.volume (%s)" % (
                                                row['name'], self.volume))
                    iblob_storid = self.injected_blob_storid
                    iblob_spath = self.injected_blob_spath
                    iblob_cksum = self.injected_blob_checksum

                DumpJob.create(cell=self.cell,
                               vl_id=row['id'],
                               server_id=rwsite['server_id'],
                               partition=rwsite['partition'],
                               injected_blob_storid=iblob_storid,
                               injected_blob_spath=iblob_spath,
                               injected_blob_checksum=iblob_cksum,
                               state='CLONE_PENDING',
                               state_descr="Waiting for clone to be scheduled")

                tx.entry_done()

        self.update(state='BRUN_CLCREATE_DONE', state_descr="Dump jobs created")

        fabs.server.made_progress()

    def sched_clones(self):
        self.update(state_descr="Scheduling volume clones")

        done = self._sched(descr="clone",
                           pend_state='CLONE_PENDING',
                           start_state='CLONE_READY')

        if done:
            self.update(state='BRUN_CLSCHED_DONE',
                        state_descr="Done running volume clones")
            fabs.server.made_progress()

        else:
            self.update(state='BRUN_CLSCHED_START',
                        state_descr="Continuing to schedule volume clones")

    def create_dumps(self):
        self.update(state_descr="Creating dump jobs")

        DumpJob.update_all(self,
                           from_state='CLONE_DONE', to_state='DUMP_PENDING',
                           timeout=0,
                           state_descr="Waiting for dump to be scheduled")

        self.update(state='BRUN_DUMPCREATE_DONE', state_descr="Dump jobs created")
        fabs.server.made_progress()

    def sched_dumps(self):
        self.update(state_descr="Scheduling volume dumps")

        done = self._sched(descr="dump",
                           pend_state='DUMP_PENDING',
                           start_state='DUMP_READY')

        if done:
            self.update(state='BRUN_DUMPSCHED_DONE',
                        state_descr="Done running volume dumps",
                        end=int(time.time()))
            fabs.server.made_progress()

        else:
            self.update(state='BRUN_DUMPSCHED_START',
                        state_descr="Continuing to run volume dumps")

    def _sched(self, descr, pend_state, start_state):
        """
        Schedule tasks for this brun (e.g. cloning volumes, dumping volumes, etc).

        Returns:
            bool: True if we are done scheduling the relevant tasks; there's no
            more cloning/dumping to do. False if there are still jobs running
            or waiting to be scheduled.
        """

        # See how many jobs are scheduled or running
        counts = DumpJob.get_counts(self, running=True)

        if not counts.total_quota_ok():
            # If we are at our total max quota, don't even try processing anything
            log.d("Skipping scheduling %s jobs, total quota at max" % descr)
            return False

        # Have we scheduled at least one new job?
        did_schedule = False

        with db.connect_bulk() as tx:
            for job in DumpJob.finditer(self, state=pend_state):
                if not counts.quota_ok(job):
                    # Skip this job
                    continue

                log.d("Scheduling %s job for vl_id %d, server_id %s partition %s" % (
                      descr, job.vl_id, job.server_id, job.partition))
                job.update(state=start_state, state_descr="About to start %s" % descr,
                           running=True)
                tx.entry_done()

                did_schedule = True

                if not counts.total_quota_ok():
                    # We've hit our total quota, so we can't schedule any more jobs
                    log.d("Not scheduling any more %s jobs; total quota full" % descr)
                    break

        if counts.total_count == 0:
            # We didn't find any NEW jobs, or any jobs that are in progress.
            # Check once more in the db just to be certain...
            pend_counts = DumpJob.get_counts(self, state=pend_state)
            run_counts = DumpJob.get_counts(self, running=True)
            if pend_counts.total_count == 0 and run_counts.total_count == 0:
                # Indeed, we cannot find any relevant jobs to run or that are
                # running. So, we are done with this stage of jobs
                # into the next state
                return True

        if did_schedule:
            log.d("made_progress in _sched")
            fabs.server.made_progress()

        # We either have more pending jobs that are waiting to be scheduled, or
        # there are some jobs still running. Either way, we are not done yet.
        return False

    def clean_cruft(self, active=False):
        """
        Clean database cruft for this brun.

        Args:
            active (bool, optional): If True, we're running this while running
                this backup run. If False (the default), make sure this brun
                isn't active, and throw an error if it is.

        Returns:
            dict: A dict indicating how many items were deleted while cleaning
            this brun. For example, {'dumpjobs': 5, 'vlentries': 3} means that
            we deleted 5 dumpjobs and 3 vlentries.

        Raises:
            `fabs.err.BrunActiveError`: This brun is actively running, and
            `active` is not True.

        """
        if not active and self.state not in ('DONE','FAILED'):
            raise err.BrunActiveError("Cannot clean active brun %d (state %s)" % (
                                      self.id, self.state))

        log.d("Cleaning cruft for brun %d" % self.id)

        ret = {}

        n_dumpjobs = db.dumpjob.clear_jobs(self)
        if n_dumpjobs > 0:
            ret['dumpjobs'] = n_dumpjobs

        n_vlentries = db.vlentry.delete_cruft(self.id)
        if n_vlentries > 0:
            ret['vlentries'] = n_vlentries

        if self.is_empty():
            self.delete()
            ret['brun'] = 1

        return ret

find = BackupRun.find
status = BackupRun.status
status_str = BackupRun.status_str
cleanable_bruns = BackupRun.cleanable_bruns

def backup_now(note, volume=None, volume_blob=None):
    return BackupRun.create(config.get('afs/cell'), note, volume, volume_blob)

def kill(brid, note):
    brun = BackupRun.find(id=brid)[0]
    brun.kill(note)

def retry(brid):
    brun = BackupRun.find(id=brid)[0]
    brun.retry()



#
# Copyright (c) 2015-2022, Sine Nomine Associates ("SNA")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND SNA DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL SNA BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

import json
import os
import pytest

import fabs.testutil as tu

import fabs.const

# Some very basic tests that don't need any db, any AFS environment, etc. Just
# run some basic commands and see that they work.

class TestBasic(tu.FabsTest):
    def test_vars(self):
        vars_ = self.fabsys_stdout('vars')

        # Just check that we got something of reasonable size in stdout
        self.comment('vars size: %d' % len(vars_))
        assert len(vars_) > 50
        assert "VERSION = " in vars_
        assert "CONF_DIR = " in vars_

    def test_db_init_sql(self):
        sql = self.fabsys_stdout('db-init --sql')
        self.comment("sql size: %d" % len(sql))
        assert len(sql) > 1000

    def test_config_check(self):
        res = self.fabsys_json('config --check')
        assert res['fabs_config_check']['ok'] is True

        # Just make sure the text version returns success; don't bother
        # checking what it sends to stdout
        self.fabsys_stdout('config --check')

    def test_config_dump_cli(self):
        cellname = 'mockcell.example.com'

        # Try setting a simple config directive via the command-line
        res = self.fabsys_json('config --dump -x afs/cell=%s' % cellname)

        assert res['fabs_config']['afs']['cell'] == cellname

    def test_config_dump_cli_xmerge(self):
        # Give several -x options, and check that they are merged properly.
        res = self.fabsys_json(('config --dump ' +
                                '-x afs/cell=example.com ' +
                                '-x afs/fs=/foo/fs ' +
                                '-x afs/aklog=/foo/aklog ' +
                                '-x afs/cell=example.net'))

        assert res['fabs_config']['afs']['cell'] == 'example.net'
        assert res['fabs_config']['afs']['fs'] == ['/foo/fs']
        assert res['fabs_config']['afs']['aklog'] == ['/foo/aklog']

    def test_config_get_fail(self):
        err = self.fabsys_stderr('config foo/bar', fail=True)
        assert "Unknown configuration directive" in err

    def test_config_get_cli(self):
        cellname = 'mockcell.example.com'
        expected = {'fabs_config': {'afs/cell': cellname}}

        # Try getting afs/cell from the command line, while overriding afs/cell
        res = self.fabsys_json('config afs/cell -x afs/cell=%s' % cellname)

        assert res == expected

    def test_config_get_text(self):
        cellname = 'mockcell.example.com'
        expected = cellname + "\n"

        res = self.fabsys_stdout('config afs/cell -x afs/cell=%s' % cellname)

        assert res == expected

    def test_config_get_complex_text(self):
        expected = '["vos"]\n'
        res = self.fabsys_stdout('config afs/vos')
        assert res == expected

    def test_config_get_complex_json(self):
        res = self.fabsys_json('config afs/vos')
        vos = res['fabs_config']['afs/vos']
        assert isinstance(vos, list)
        assert vos == ['vos']

    def test_config_get_format(self):
        cellname = 'mockcell.example.com'
        expected = {'fabs_config': {'afs/cell': cellname}}

        # fabsys_json will append a final '--format json', which should
        # override our '--format txt' here.
        res = self.fabsys_json('config afs/cell -x afs/cell=%s --format txt' % cellname)
        assert res == expected

    def test_config_dump_file(self):
        cellname = 'mockcell.example.com'

        # Try setting the config via a config file
        with self.mkftemp('config') as config_fh:
            self.comment('config file: %s' % config_fh.name)
            config_fh.write("afs:\n  cell: '%s'\n" % cellname)

            res = self.fabsys_json('config --dump --config %s' % config_fh.name)

        assert res['fabs_config']['afs']['cell'] == cellname

    def test_config_meta(self):
        bufsize = 1234

        res = self.fabsys_json('config --dump -x bufsize=%s' % bufsize)
        assert res['fabs_config']['bufsize'] == bufsize

        # Make sure that config stuff isn't persisting between fabsys runs
        res = self.fabsys_json('config --dump')
        assert 'bufsize' not in res['fabs_config']

    def test_version(self):
        out = self.fabsys_stdout('--version').rstrip()
        assert out == ('fabs %s' % fabs.const.VERSION)



#
# Copyright (c) 2015-2022, Sine Nomine Associates ("SNA")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND SNA DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL SNA BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

import time

import fabs.path
import fabs.util as util
import fabs.err as err
import fabs.db as db
import fabs.bstore as bstore
import fabs.log
log = fabs.log.getLogger(__name__)

class VolDump:
    """
    A VolDump represents a dump from a volume during a backup run.

    Don't instantiate a VolDump directly. Instead, you can find existing ones
    with `find`, and related functions.
    """

    @classmethod
    def find_latest(cls, *args, **kwargs):
        """
        Returns:
            `VolDump`: The latest voldump matching the given search criteria.
        """
        row = db.voldump.find_latest(*args, **kwargs)
        if row is None:
            return None
        return VolDump(row)

    @classmethod
    def find_oldest(cls, *args, **kwargs):
        """
        Returns:
            `VolDump`: The oldest voldump matching the given search criteria.
        """
        row = db.voldump.find_oldest(*args, **kwargs)
        if row is None:
            return None
        return VolDump(row)

    @classmethod
    def find(cls, **kwargs):
        """
        Returns:
            list of `VolDump`: Volume dumps matching the given search criteria.
        """
        rows = db.voldump.find(**kwargs)

        # If we were given an id, the voldump should really exist
        if 'id' in kwargs and not rows:
            raise err.BadVoldumpIdError("Voldump %d does not seem to exist" % kwargs['id'])

        return [VolDump(row) for row in rows]

    @classmethod
    def find_byblob(cls, cell, dblob):
        """
        Find the volume dump for the given dump blob and cell.
        """
        vdumps = cls.find(cell=cell,
                          dump_storid=dblob.storid(),
                          dump_spath=dblob.rel_path())
        if len(vdumps) == 0:
            return None
        return vdumps[0]

    @classmethod
    def search(cls, cell=None, brid=None, rwid=None, before_ts=None,
               after_ts=None, redundant=None, n_max=0):
        """
        Search for volume dumps, for administrative/maintenance operations.
        """

        for row in db.voldump.finditer(cell=cell, brid=brid, rwid=rwid,
                                       before_ts=before_ts, after_ts=after_ts,
                                       redundant=redundant):
            yield VolDump(row)
            if n_max > 0:
                n_max -= 1
                if n_max == 0:
                    return

    @classmethod
    def search_user(cls, cell, user, path, vol_rwid, timestamp, max_dumps):
        """
        Similar to `search`, but intended more for looking for volume dumps to
        restore data from.
        """

        if path and vol_rwid:
            raise err.InternalError("VolDump.search_user cannot handle both path and volume")
        if not path and not vol_rwid:
            raise err.InternalError("VolDump.search_user must have either path or volume")

        if vol_rwid:
            if user:
                util.check_restore_authz(user, cell, vol_rwid)
            vd_list = cls._search_vol(cell, vol_rwid, timestamp, max_dumps)

        else:
            vd_list = cls._search_path(path, timestamp, max_dumps)
            if user:
                check_vols = {}
                for voldump in vd_list:
                    check_vols[voldump.rwid] = True

                for rwid in check_vols.keys():
                    util.check_restore_authz(user, cell, rwid)

        return vd_list

    @classmethod
    def _search_vol(cls, cell, rwid, timestamp, max_dumps):
        # First, find one volume dump to look at
        search_after = False

        # Hmm, should 'rwid' just be a numeric volume id, or should we maybe be
        # able to search by volume name? For now we just do numerical id
        # lookups, but we could possibly look volumes up by name in the db.
        # However, this possibly has security ramifications, in case volume
        # names changed or something (we only do the authz check once, and if
        # the volid changed, we would be checking a different volume...). Not
        # sure on the right behavior there, so just do numeric ids for now.

        if timestamp is None:
            timestamp = int(time.time())

        first_vd = cls.find_oldest(cell, rwid=rwid, after_time=timestamp)
        if first_vd is None:
            search_after = True
            first_vd = cls.find_latest(cell, rwid=rwid, before_time=timestamp)

        if first_vd is None:
            # We can't find any volume dump
            return []

        oldest_vd = first_vd
        latest_vd = first_vd
        vd_list = [first_vd]

        last_chance = False

        log.d("Looking for a max of %d voldumps" % max_dumps)

        # Now that we have one volume dump, look for others adjacent to it,
        # until we have max_dumps volume dumps
        while len(vd_list) < max_dumps:
            log.d("Got %d voldumps, looking for %d" % (len(vd_list), max_dumps))

            if search_after:
                if not last_chance:
                    # Alternate between directions, unless we ran out of
                    # matches in one direction (and thus 'last_chance' is set)
                    search_after = False
                next_vd = cls.find_oldest(cell, rwid=rwid, after_id=latest_vd.id)
                if next_vd is not None:
                    vd_list.append(next_vd)
                    latest_vd = next_vd

            else:
                if not last_chance:
                    search_after = True
                next_vd = cls.find_latest(cell, rwid=rwid, before_id=oldest_vd.id)
                if next_vd is not None:
                    # Prepend to the vd_list
                    vd_list[:0] = [next_vd]
                    oldest_vd = next_vd

            if next_vd is None:
                # We found no more volume dumps in this direction, so try the
                # other direction before giving up.
                if last_chance:
                    # ...but we already are trying the "other" direction. So,
                    # give up.
                    break

                log.d("Found no more voldumps, looking only %s from now on" % (
                      "forwards" if search_after else "backwards"))

                last_chance = True

        return vd_list

    @classmethod
    def _search_path(cls, path, timestamp, max_dumps):
        # First just use the translator to get a volume id from our path
        xlate = fabs.path.PathTranslator(path, timestamp)
        voldump = xlate.translate()

        # Then, just use the normal algorithm for going from a volume id to a
        # list of vol dumps
        return cls._search_vol(voldump.cell, voldump.rwid, timestamp, max_dumps)

    def __init__(self, row):
        self._cols = list(row.keys())
        for col in self._cols:
            setattr(self, col, row[col])

    def get_brun(self):
        return fabs.brun.find(id=self.br_id)[0]

    def delete(self, keep_blob):
        blob = None

        with db.connect():
            brun = self.get_brun()
            if brun.state not in ('DONE', 'FAILED'):
                raise err.VolDumpActiveError(
                        ("Cannot delete voldump %d from active backup run " +
                         "%d (state %s)") % (
                         self.id, brun.id, brun.state))

            if not keep_blob:
                blob = self.dump_blob()

            fabs.restore.clear_voldump(self)
            db.voldump.delete(self.id)
            db.links.clear(self.vl_id)
            db.vlentry.delete(self.vl_id)

        if blob is not None:
            blob.delete(suppress_error=FileNotFoundError)

    def as_dict(self):
        ret = {}
        for attr in self._cols:
            ret[attr] = getattr(self, attr)
        return ret

    def describe_dict(self):
        vdinfo = self.as_dict()

        vdinfo['br_id'] = db.vlentry.get(vdinfo['vl_id'])['br_id']
        vdinfo['dump_blob'] = self.dump_blob_dict()

        return vdinfo

    def describe(self):
        vdinfo = self.describe_dict()

        update_secs = vdinfo['hdr_update']
        if update_secs is None or update_secs == 0:
            update_ctime = 'Never'
        else:
            update_ctime = time.ctime(update_secs)
        vdinfo['update_ctime'] = update_ctime

        creation_secs = vdinfo['hdr_creation']
        if creation_secs is None or creation_secs == 0:
            creation_ctime = 'Never'
        else:
            creation_ctime = time.ctime(creation_secs)
        vdinfo['creation_ctime'] = creation_ctime

        dump_path = None
        if vdinfo['dump_blob'] is not None:
            dump_path = vdinfo['dump_blob']['abs_path']
        if dump_path is None:
            dump_path = '[none]'
        vdinfo['dump_path'] = dump_path

        return ("Volume dump id %(id)d, volume: %(rwid)d (%(name)s)\n" +
                "  cloned at:    %(creation_ctime)s\n" +
                "  last updated: %(update_ctime)s\n" +
                "  dump blob: %(dump_path)s") % vdinfo

    def dump_blob(self):
        return bstore.blob_from_db(self.dump_storid, self.dump_spath, checksum=self.dump_checksum)

    def dump_blob_dict(self):
        # Note: This is best-effort; failures to resolve the dump blob in the
        # db will just log a warning. This is just for CLI frontend stuff, not
        # for actual dump manipulation.
        return bstore.blobdict_from_db(self.dump_storid, self.dump_spath, checksum=self.dump_checksum)

find_latest = VolDump.find_latest
find_oldest = VolDump.find_oldest
find_byblob = VolDump.find_byblob
find = VolDump.find
search = VolDump.search
search_user = VolDump.search_user



#
# Copyright (c) 2015-2022, Sine Nomine Associates ("SNA")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND SNA DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL SNA BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

import multiprocessing
import signal

import fabs.err as err
import fabs.brun
import fabs.dumpjob
import fabs.restore
import fabs.util
import fabs.const
import fabs.log
log = fabs.log.getLogger(__name__)

do_quit = False

def _int_handler(signum, frame):
    # pylint: disable=unused-argument
    global do_quit
    do_quit = True

def run(once=False):
    """
    Run the server main loop.

    Args:
        once (bool, optional): If True, just run the main loop exactly once,
            then return. Otherwise, run forever until interrupted.
    """

    log.info('start', "Starting fabs server v%s" % fabs.const.VERSION)
    int_sleep.setup()

    if not once:
        signal.signal(signal.SIGINT, _int_handler)
        signal.signal(signal.SIGTERM, _int_handler)

    while not do_quit:
        try:
            fabs.brun.BackupRun.process_jobs()
        except Exception:
            log.exception('bruns_err', "Error while processing backup runs")

        try:
            fabs.dumpjob.DumpJob.process_jobs()
        except Exception:
            log.exception('dumpjobs_err', "Error while processing dump jobs")

        try:
            fabs.restore.RestoreRequest.process_jobs()
        except Exception:
            log.exception('restore_err', "Error while processing restore requests")

        if once:
            break

        int_sleep.sleep(60)

    log.info('stop', "Fabs server shutting down")

def made_progress():
    """
    In a child process, indicate that we have made some forward process.

    This wakes the parent process (the server process), so it will immediately
    check the db to see if there is more work to do.
    """
    int_sleep.wake_parent()

class _InterruptibleSleep:
    """
    Lets the caller sleep for a certain amount of time, but can be woken up
    early by child processes calling `made_progress`.
    """

    def __init__(self):
        self.event = None

    def setup(self):
        if self.event is not None:
            # We've already been setup
            return
        self.event = multiprocessing.Event()

    def wake_parent(self):
        if self.event is None:
            raise err.InternalError("int sleep event is not set, but we called wake_parent")
        self.event.set()

    def sleep(self, seconds):
        """
        Sleep for approximately the given number of seconds, unless woken up
        early.

        Args:
            seconds (int): The number of seconds to sleep.
        """

        if not self.event.is_set():
            log.d("Waiting up to %d seconds for an event to trigger" % seconds)

        # Wait to see if self.event is triggered, but we also want to check if
        # do_quit is set. Python has various issues with triggering an Event in
        # a signal handler, so sidestep the whole issue and check out do_quit
        # flag here.
        for _ in range(seconds):
            self.event.wait(1)
            if self.event.is_set():
                break
            if do_quit:
                break

        # Periodically clean up any lingering zombies created from FabsProcess
        # jobs
        # pylint: disable=not-callable
        multiprocessing.active_children()

        self.event.clear()

int_sleep = _InterruptibleSleep()



#
# Copyright (c) 2015-2022, Sine Nomine Associates ("SNA")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND SNA DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL SNA BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

import contextlib
import os
import os.path
import time
import hashlib

import fabs.server
import fabs.err as err
import fabs.voldump as voldump
import fabs.bstore as bstore
import fabs.util as util
import fabs.db as db
import fabs.config as config
import fabs.log
log = fabs.log.getLogger(__name__)

class DumpJob:
    """
    A DumpJob represents a task that handles dumping a volume to disk.

    Don't instantiate a DumpJob directly. Instead, create a new job with
    `create`, or find existing jobs with `find`.
    """

    # If a volume appears to have changed less than FUDGE_TIME seconds since
    # before the last backup, we assume that backup didn't catch it, so we
    # assume the volume has changed, even though it looks like it hasn't
    # changed since the last backup. We do this just to try to be on the safe
    # side.
    FUDGE_TIME = 3

    @classmethod
    def finditer(cls, brun, **kwargs):
        """
        Find dump jobs for the given brun and given criteria.

        Args:
            `fabs.brun.BackupRun`: The backup run to search in.
            **kwargs: Search criteria. See `db.dumpjob.finditer`.

        Yields:
            `DumpJob`: Dump jobs that match the given criteria.
        """
        for row in db.dumpjob.finditer(brun, **kwargs):
            yield DumpJob(row)

    @classmethod
    def find(cls, **kwargs):
        """
        Find dump jobs that match the given criteria.

        Args:
            **kwargs: Search criteria. See `db.dumpjob.find`.

        Returns:
            list of `DumpJob`: Matching dump jobs.
        """
        rows = db.dumpjob.find(**kwargs)
        return [DumpJob(row) for row in rows]

    @classmethod
    def describe(cls, brun, **kwargs):
        jobinfos = db.dumpjob.describe(brun, **kwargs)
        return jobinfos

    @classmethod
    def get_counts(cls, brun, **kwargs):
        """
        Get how many dump jobs match the given criteria.

        Args:
            `fabs.brun.BackupRun`: Backup run to search in.
            **kwargs: Search criteria. See `db.dumpjob.get_counts`.

        Returns:
            `QuotaCounts`: The counts for the matching jobs (contains a total
            count, per-server, and per-partition counts).
        """

        part_counts = db.dumpjob.get_counts(brun, **kwargs)

        srv_counts = {}
        total_count = 0

        for ((srv, _), count) in part_counts.items():
            srv_counts[srv] = srv_counts.get(srv, 0) + count
            total_count += srv_counts[srv]

        return QuotaCounts(total_count, srv_counts, part_counts)

    @classmethod
    def clear(cls, brun):
        """
        Delete all dumpjobs for the given brun.
        """
        db.dumpjob.clear(brun)

    @classmethod
    def create(cls, *args, **kwargs):
        db.dumpjob.create(*args, **kwargs)

    @classmethod
    def find_update(cls, cell, from_state, to_state, state_descr, **kwargs):
        """
        Transition any matching dump jobs in `from_state` to `to_state`, and
        set their status description to `state_descr`, andupdate any fields in
        **kwargs.
        """
        update = dict(state=to_state)
        if state_descr is not None:
            update['state_descr'] = state_descr
        update.update(kwargs)

        db.dumpjob.update_all(match=dict(cell=cell,
                                         state=from_state),
                              update=update)

        return cls.find(cell=cell, state=to_state)

    @classmethod
    def update_all(cls, brun, from_state, to_state, **kwargs):
        db.dumpjob.update_all(match=dict(br_id=brun.id,
                                         state=from_state),
                              update=dict(state=to_state,
                                          **kwargs))

    @classmethod
    def process_jobs(cls):
        cell = config.get('afs/cell')

        # Create our backup clone
        cls._process(cell, 'CLONE_READY', 'CLONE_START', 'CLONE_WORK',
                     cls.do_clone, "Starting volume clone")

        # Do our 'vos dump'
        cls._process(cell, 'DUMP_READY', 'DUMP_START', 'DUMP_WORK',
                     cls.do_dump, "Starting volume dump")

        # Don't use _process for this, since we do not want to find or do
        # anything with the resultant jobs in the DONE state.
        db.dumpjob.update_all(match=dict(cell=cell,
                                         state='DUMP_DONE'),
                              update=dict(state='DONE',
                                          state_descr="Done with volume",
                                          timeout=0))

        db.dumpjob.update_all(match=dict(cell=cell,
                                         state='DUMP_SKIPPED'),
                              update=dict(state='DONE',
                                          timeout=0))

        cls._check_stale()

        # Retry previous stages that failed with an error
        cls._process(cell, 'ERROR', 'DUMP_RECOVER_START', 'DUMP_RECOVER_WORK',
                     cls.do_recover, None)

    # todo: This is very similar to the brun._process method; could be
    # consolidated...
    @classmethod
    def _process(cls, cell, from_state, to_state, work_state, func, descr):
        jobs = cls.find_update(cell, from_state, to_state, descr, timeout=0)
        log.d("Found %d jobs from %s -> %s" % (len(jobs), from_state, to_state))

        if not jobs:
            # no matching jobs; nothing to do
            return

        procs = []

        with db.connect_bulk() as tx:
            for job in jobs:
                if work_state is not None:
                    job.update(state=work_state)
                    tx.entry_done()

                if func is not None:
                    procs.append(util.FabsProcess(target=job.process_child, args=(func,)))

        for proc in procs:
            proc.start()

    @classmethod
    def _check_stale(cls):
        jobs = cls.find(stale=True)
        if not jobs:
            return

        with db.connect_bulk() as tx:
            for job in jobs:
                log.error('job_stale', ("%s has taken too long to make " +
                          "progress. It may be stuck, or exited uncleanly. " +
                          "FABS will mark this as an error and attempt to " +
                          "retry it. (%s)") % (job, job.details()))
                job.error()
                tx.entry_done()

    def __init__(self, row):
        for col in row.keys():
            setattr(self, col, row[col])

    def __str__(self):
        return "dump job vl_id %s (vol %d, %s)" % (self.vl_id, self.rwid, \
                                                   self.name)

    def details(self):
        return "state %s mtime %d timeout %d descr %s" % (self.state,
                self.mtime, self.timeout, self.state_descr)

    def update(self, **kwargs):
        db.dumpjob.update(self, **kwargs)

    def process_child(self, func):
        try:
            func(self)
        except Exception:
            log.exception('child_err',
                          "Error when processing %s" % self)
            self.error()

    def error_nonfatal(self):
        db.dumpjob.error_nonfatal(self)

    def error(self):
        """
        Flag this dumpjob as encountering an error.
        """
        if self.state in ('ERROR', 'DUMP_RECOVER_START', 'DUMP_RECOVER_WORK'):
            log.error('dump_err_meta', ("%s encountered an error " +
                      "while trying to recover from a previous error. It " +
                      "will not be retried.") % self)
            self.update(state='FAILED', running=False, timeout=0)

        else:
            db.dumpjob.error(self)

    # If the volume should be skipped, return the skip reason (a string like
    # 'UNCHANGED'). If the volume should NOT be skipped, return None.
    @classmethod
    def should_skip(cls, cell, volname, rwid=None, dumpjob=None):
        """
        Determine if the given volume should be skipped.

        Args:
            cell (str): Cell for the volume.
            volname (str): Name of the volume.
            rwid (int, optional): RW volid of the volume, if known.
            dumpjob (`DumpJob`, optional): The associated dumpjob, if any, for
                logging purposes.

        Returns:
            str or None: If the volume should be skipped, return a skip reason
            (e.g. 'UNCHANGED'). If the volume should NOT be skipped, return
            None.
        """

        vos = util.vos_unauth(cell=cell)

        if rwid is None:
            res = vos.listvldb(name=volname, noresolv=True).run()
            for entry in res.entries:
                rwid = entry.rwid

        res = vos.examine(id=rwid, noresolv=True).run()
        cur_update_time = res.headers[0].update_time

        if res.headers[0].status is fabs.cmd.VolumeHeader.offline:
            dumpjob_str = ''
            if dumpjob is not None:
                dumpjob_str = ' (%s)' % dumpjob

            if_offline = config.get('dump/if_offline')
            if if_offline == 'skip':
                log.info('offline_skip', ("Skipping offline volume %d%s. To " +
                         "change this behavior, see option dump/if_offline.") % (rwid, dumpjob_str))
                return 'OFFLINE'

            else:
                assert if_offline == 'error'
                raise err.DumpOfflineError(("Volume %d is offline, and we cannot " +
                                            "backup offline volumes. See option dump/if_offline to " +
                                            "skip offline volumes instead.%s") % (rwid, dumpjob_str))


        last_voldump = voldump.find_latest(cell=cell, rwid=rwid)
        if last_voldump is None:
            log.d("No previous dump for %s, volume must be dumped" % volname)
            return None

        if last_voldump.hdr_update != cur_update_time:
            # The volume has clearly changed, so we must dump it
            log.d("Volume %s has obviously changed (%d != %d)" % (
                  volname, last_voldump.hdr_update, cur_update_time))
            return None

        # The time that the backup volume was created is recorded in both
        # the 'creation' and 'backup' volume header fields. Use the oldest one
        # of those, just in case they are different for any reason.
        backup_ran = min(last_voldump.hdr_backup, last_voldump.hdr_creation)

        if backup_ran < last_voldump.hdr_update:
            # According to these timestamps, the last backup was run before the
            # last update was made to the volume. That doesn't make any sense;
            # err on the side of backing up too much.

            log.warn('lastdump_time', ("Volume %s (voldump %d) claims its backup " +
                     "clone was created around timestamp %d (backup %d, " +
                     "creation %d), but was last updated around timestamp " +
                     "%d. That does not make much sense, so we are assuming " +
                     "this volume needs to be backed up.") % (
                     volname, last_voldump.id, backup_ran, last_voldump.hdr_backup,
                     last_voldump.hdr_creation, last_voldump.hdr_update))
            return None

        if backup_ran - last_voldump.hdr_update < cls.FUDGE_TIME:
            log.d(("volume %s appears to not have changed, but its " +
                   "hdr_update (%d) is close enough to the backup time " +
                   "(%d) that we're backing it up anyway") % (
                   volname, last_voldump.hdr_update, backup_ran))
            return None

        max_age = config.get('dump/max_unchanged_age')
        if max_age:
            curtime = int(time.time())
            age = curtime - backup_ran
            if backup_ran < curtime and age > max_age:
                # Force a new dump if the last dump for this vol is more than
                # max_unchanged_age seconds old
                log.d(("The last backup for volume %s is %d seconds old, " +
                       "which is greater than dump/max_unchanged_age (%d). " +
                       "forcing a new dump.") % (volname, age,
                       max_age))
                return None

        # The volume "Last Update" time has not changed since the last backup
        # run, and the "Last Update" time is far away from the last time a
        # backup was created. We can pretty confidently say that the volume
        # has not changed, so we do not need to back it up again.
        log.d("Volume %s appears to not have changed since last backup" % volname)
        return 'UNCHANGED'

    def do_clone(self):
        vlentry = db.vlentry.get(self.vl_id)

        vos = util.vos_unauth(cell=self.cell)

        self.update(state_descr="Checking if volume has changed since last backup")

        skip_reason = self.should_skip(self.cell, vlentry['name'], vlentry['rwid'])
        if skip_reason is None:
            self.update(state_descr="Cloning volume %s" % vlentry['name'])

            vos = util.vos_auth(cell=self.cell)
            cursor = vos.backup(id=vlentry['name'])

            log.d("Running vos backup for %s in pid %d" % (
                  vlentry['name'], cursor.get_proc().pid))

            cursor.run()

            log.d("Looking up new vldb info for volume")

            res = vos.listvldb(name=vlentry['name']).run()
            db.vlentry.update(self.vl_id, res.entries[0])

            self.update(state='CLONE_DONE',
                        state_descr="Done cloning volume",
                        running=False,
                        timeout=0)

        else:
            self.update(state='DUMP_SKIPPED',
                        state_descr="Skipping backup because volume is '%s'" % skip_reason,
                        running=False,
                        skip_reason=skip_reason,
                        timeout=0)

        fabs.server.made_progress()

    def _header_match(self, volname, vos_header, dump_path):
        """
        Returns whether the header in the given dump file matches the header of
        the live volume.
        """
        dumpscan = util.dumpscan()

        self.update(state_descr="Running dumpscan to examine vol header")

        file_header = dumpscan.scan_volheader(dump_path).run()

        # Compare these attributes between the 'vos' header, and the header
        # from the on-disk header that we just got from dumpscan.
        for attr in ('rwid', 'update_time'):
            vos_val = getattr(vos_header, attr)
            file_val = getattr(file_header, attr)

            if vos_val != file_val:
                log.warn('hdr_mismatch', ("'vos examine' said that the value " +
                         "for '%s' for volume %s is %r, but the value in the " +
                         "dump blob we just got was %r") % (attr, volname,
                         vos_val, file_val))
                return False

        return True

    def injected_blob(self):
        return bstore.blob_from_db(self.injected_blob_storid,
                                   self.injected_blob_spath,
                                   self.injected_blob_checksum)

    def do_dump(self):
        # Dummy class for a special exception case below
        class SkipDump(BaseException): pass

        vlentry = db.vlentry.get(self.vl_id)
        vos = util.vos_auth(cell=self.cell)

        self.update(state_descr="Examining volid %s" % vlentry['bkid'])

        res = vos.examine(id=vlentry['bkid']).run()
        header = res.headers[0]

        self.update(state_descr="Dumping volume %s" % vlentry['name'])

        proc = None
        dblob = None
        skipped = False

        try:
            injected_blob = self.injected_blob()
            if injected_blob:
                # We already have a blob to use
                log.d(("Skipping vos dump for vol %s, since we were given a " +
                       "dump to inject") % vlentry['name'])
                dblob = injected_blob
                dblob.set_tmp()

            else:
                # We don't have a blob; get one from 'vos dump'

                self.update(state_descr="Getting volume %s size" % vlentry['name'])

                dump_size = vos.size(id=vlentry['bkid'], dump=True).run()

                self.update(state_descr="About to alloc new blob for volume dump")
                dblob = bstore.BlobStore.alloc_blob(dump_size,
                                                    "vosdump_%d_%d" % (vlentry['rwid'], self.vl_id),
                                                    update=self.update)

                cursor = vos.dump(id=vlentry['bkid'])
                proc = cursor.get_proc()

                log.d("Running vos dump %d in pid %d" % (vlentry['bkid'], proc.pid))

                self._handle_vosdump(cursor, dump_size, dblob)

                proc = None

            # Check if the header.update_time et al match what's in the actual
            # dump blob we have
            if not self._header_match(vlentry['name'], header, dblob.abs_path()):
                if self.injected_blob():
                    # For injected dumps, a header mismatch is okay; we just
                    # ignore the dump
                    log.info('inject_stale', ("The injected dump for volume " +
                             "%s appears to not match the live volume. " +
                             "Ignoring the injected dump.") % vlentry['name'])
                    raise SkipDump()

                raise err.DumpRaceError(("Dump blob header does not match " +
                                         "'vos examine' info for volume %s (%s)") % (
                                         vlentry['bkid'], self))

            # Store symlinks in the db
            self._handle_links(dblob)

            # Calculate our checksum
            self.update(state_descr="Verifying checksum")
            dblob.verify_checksum()

            self.update(state_descr="Committing completed dump")

            # Get the real dump file size on disk
            dump_size = os.path.getsize(dblob.abs_path())

            # Move the dump blob to a permanent, non-tmp path
            dblob.commit(self, vlentry['rwid'])

            db.voldump.create(vl_id=self.vl_id,
                              hdr_size=header.size,
                              hdr_creation=header.create_time,
                              hdr_copy=header.copy_time,
                              hdr_backup=header.backup_time,
                              hdr_update=header.update_time,
                              dump_size=dump_size,
                              dump_storid=dblob.storid(),
                              dump_spath=dblob.rel_path(),
                              dump_checksum=dblob.checksum())

        except SkipDump:
            skipped = True

        except Exception as exc:
            try:
                if dblob:
                    dblob.cleanup()

                # Try to kill the underlying process, before we bail out
                if proc and proc.poll() != None:
                    # Ignore ESRCH errors; that'll just happen sometimes, if
                    # the process dies as we're trying to kill it.
                    with contextlib.suppress(ProcessLookupError):
                        proc.kill()

            except Exception:
                # Log any errors here to not ignore them, but still throw the
                # original 'exc' that got us here.
                log.exception('vosdump_cleanu_p_err',
                              "Error when trying to cleanup vos process after other error")
            finally:
                raise exc

        if skipped:
            # If we're skipping this volume, we're not going to be using the
            # dump blob that we have, so get rid of it.
            dblob.delete()

            self.update(state='DUMP_SKIPPED',
                        state_descr="Skipping dump of volume due to stale injected dump blob",
                        injected_blob_storid=None,
                        injected_blob_spath=None,
                        injected_blob_checksum=None,
                        running=False,
                        skip_reason='STALE_INJECT',
                        timeout=0)
        else:
            self.update(state='DUMP_DONE',
                        state_descr='Done dumping volume',
                        injected_blob_storid=None,
                        injected_blob_spath=None,
                        injected_blob_checksum=None,
                        running=False,
                        timeout=0)
        fabs.server.made_progress()

    def _handle_vosdump(self, cursor, dump_size, dblob):
        cursor.steal_stdout()

        ck_algo = config.get('dump/checksum')
        ck_m = hashlib.new(ck_algo)

        bufsize = config.get('bufsize')
        nbytes = 0
        pretty = util.PrettyBytes(dump_size)

        # Update status every 2 minutes
        update_int = 60*2

        # Fudge the initial last_update, so our first update happens in about
        # 10 seconds, but then 'update_int' seconds after that
        last_update = int(time.time()) - update_int + 10

        # 10 minute timeout
        read_timeout = 60 * 10

        with open(dblob.abs_path(), 'r+b') as dump_fh:
            while True:
                buf = cursor.read_stdout(bufsize, timeout=read_timeout)
                if buf is None:
                    # EOF
                    break

                ck_m.update(buf)
                dump_fh.write(buf)

                nbytes += len(buf)

                now = int(time.time())
                if now - last_update > update_int:
                    pretty.update_bytes(nbytes)
                    self.update(timeout=read_timeout*2,
                                state_descr="Running vos dump (%s / %s dumped, %s)" % (
                                             pretty.bytes, pretty.total, pretty.rate))
                    last_update = now

            dump_fh.flush()

            filesize = os.fstat(dump_fh.fileno()).st_size
            if filesize < nbytes:
                raise err.InternalError("Dumped file too small? %d < %d" % (
                                        filesize, nbytes))
            # Truncate the file back to size; it may be larger because of space
            # that was reserved pre-dump
            dump_fh.truncate(nbytes)

        dblob.set_checksum('%s:%s' % (ck_algo, ck_m.hexdigest()))
        self.update(state_descr="Done vos dumping volume")

    def _handle_links(self, dblob):
        self.update(state_descr="Running dumpscan to find symlinks")

        try:
            db.links.clear(self.vl_id)

            dumpscan = util.dumpscan()
            links = dumpscan.scan_symlinks(dblob.abs_path()).run()

            if links:
                with db.connect():
                    self.update(state_descr="Adding links to db")
                    db.links.add(self.vl_id, links)

        except Exception as exc:
            # Something went wrong when scanning for symlinks. What should we
            # do? Technically this is not required to backup the volume, but it
            # can confusingly break our path processing for restores. So, we
            # leave the decision to a config option.

            log.error('scan_err', "Error when scanning symlinks for %s" % self)

            if config.get('dump/scan_links/abort_on_error'):
                log.error('scan_err', ("The dump for this volume is being " +
                          "aborted. To change this behavior, see the option " +
                          "dump/scan_links/abort_on_error"))
                raise exc

            log.error('scan_err', ("Continuing to backup this volume, but " +
                      "path information for the volume will not be stored. " +
                      "To change this behavior, see the option " +
                      "dump/scan_links/abort_on_error."))
            log.exception('scan_err', "Error:")

            self.error_nonfatal()

        self.update(state_descr="Done scanning links")

    def do_recover(self):
        # e.g. If a state failed in CLONE_WORK, put it back in CLONE_START
        ret_table = dict(
            CLONE_WORK='CLONE_START',
            DUMP_WORK='DUMP_START',
        )

        failed = False

        if self.errors >= config.get('dump/error_limit'):
            log.error('dump_err_limit', ("%s has failed %d " +
                      "times, which exceeds our retry limit") % (
                      self, self.errors))
            failed = True

        elif self.state_last not in ret_table:
            log.error('dump_err_unknown', ("%s encountered " +
                      "an error in state %s, which we do not know how to " +
                      "retry (%s)") % (self, self.state_last, self.details()))
            failed = True

        if failed:
            self.update(state='FAILED', running=False, timeout=0)
        else:
            retry_state = ret_table[self.state_last]
            log.warn('dump_retry', "Retrying %s %s -> %s" % (
                     self, self.state_last, retry_state))
            self.update(state=retry_state, state_last=None, timeout=0)

        fabs.server.made_progress()

class QuotaCounts:
    """
    A QuotaCounts object contains the count of how many DumpJobs are running in
    total, for each server, and for each partition. A QuotaObject is used to
    calculate if any of these counts exceed their relevant quota (we may have a
    quota for the total number of dump jobs, as well as per-server and
    per-partition quotas).
    """
    def __init__(self, total_count, srv_counts, part_counts):
        self.total_count = total_count
        self.srv_counts = srv_counts
        self.part_counts = part_counts

    def quota_ok(self, job):
        """
        Would running the given job put us over any quota?

        Args:
            `DumpJob`: The job to check.

        Returns:
            bool: True if it's okay to run the given job. False if running the
            given job would put us over quota; we need to wait.
        """
        part_count = self.part_counts.get((job.server_id, job.partition), 0)
        if part_count >= config.get('dump/parallel/partition'):
            log.d("Partition quota maxed for %s server %d partition %s" % (
                  job, job.server_id, job.partition))
            return False

        srv_count = self.srv_counts.get(job.server_id, 0)
        if srv_count >= config.get('dump/parallel/server'):
            log.d("Server quota maxed for %s server %d partition %s" % (
                  job, job.server_id, job.partition))
            return False

        # Callers should check this on their own, to avoid checking this
        # per-job, but just in case...
        if not self.total_quota_ok():
            log.d("Total quota maxed for %s server %d partition %s" % (
                  job, job.server_id, job.partition))
            return False

        # Okay, we appear to be below all quotas. Increase our usage counts, to
        # indicate our resource usage for this job.
        self.srv_counts[job.server_id] = srv_count + 1
        self.part_counts[(job.server_id, job.partition)] = part_count + 1
        self.total_count = self.total_count + 1

        return True

    def total_quota_ok(self):
        """
        Are we over the total overall dumpjob quota?

        This is different than `quota_ok`, because if this returns True, we are
        over quota for _any_ job.

        Returns:
            bool: True if running any new job would put us over quota. False
            otherwise.
        """
        if self.total_count >= config.get('dump/parallel/total'):
            return False
        return True



#
# Copyright (c) 2015-2022, Sine Nomine Associates ("SNA")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND SNA DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL SNA BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

import time

import fabs.voldump
import fabs.err as err
import fabs.db as db
import fabs.log
log = fabs.log.getLogger(__name__)

class PathTranslator:
    """
    A PathTranslator is used to translate a plain string path (and timestamp)
    into a specific fabs volume dump. Example usage:

        xlator = PathTranslator('/afs/foo/bar/baz', int(time.time()))
        voldump = xlator.translate()
    """

    # A very high limit, just to prevent endless loops.
    # This is way higher than most symlink limits in actual OS VFSes, so we
    # shouldn't encounter any false positives with this.
    MAX_SYMLINKS = 1000

    def __init__(self, path, timestamp):
        """
        Args:
            path (str): Path to translate.
            timestamp (int): When we should go back in time to translate the
                given path, as a unix timestamp.
        """
        if timestamp is None:
            timestamp = int(time.time())

        self._timestamp = timestamp
        self._orig_path = path
        self._n_symlinks = 0

        if not path.startswith('/afs/'):
            raise err.PathError(("The provided path '%s' does not " +
                                 "start with /afs/") % path)
        self._path = self._split_path(path)
        self._path_pos = 0
        self._voldump = None

        log.d("Starting path translation for path '%s' around timestamp %d" % (
              path, timestamp))

    @classmethod
    def _split_path(cls, path):
        # Given a path like foo/bar/baz, split into [foo, bar, baz]
        return [comp for comp in path.split('/') if len(comp) > 0]

    def translate(self):
        """
        Returns:
            `fabs.voldump.VolDump`: The closest volume dump we can find
            corresponding to the given path.
        """
        while not self._done():
            self._traverse_one()
        return self._voldump

    def _done(self):
        # If we cannot pull off any more elements from _path, we have nothing
        # more to do, so we're done
        if self._path_pos >= len(self._path):
            return True
        return False

    def path_str(self):
        return '/'.join([''] + self._path)

    def _pop_path(self):
        if self._done():
            raise err.PathError(("Expected another path element at index %d " +
                                 "for path '%s'") % (
                                 self._path_pos, self.path_str()))

        ret = self._path[self._path_pos]
        self._path_pos += 1
        return ret

    def _traverse_one(self):
        if self._path_pos == 0:
            # We're starting from the root
            self._voldump = None
            self._flatten_dots()

            first = self._pop_path()
            if first != 'afs':
                raise err.PathError("Unknown root dir '%s' in path '%s'" % (
                                    first, self.path_str()))

            cellname = self._pop_path()
            if cellname.startswith('.'):
                # Trim off leading '.', if this is just an RW mount
                cellname = cellname[1:]

            self._set_vol(cellname, 'root.cell')
            return

        # Keep track of our path relative to the volume root
        cur_path = ''

        while not self._done():
            dentry = self._pop_path()

            cur_path += '/'
            cur_path += dentry

            links = db.links.search(vl_id=self._voldump.vl_id,
                                    path=cur_path)
            if links:
                # pylint: disable=unused-variable
                (path, target) = links[0]
                self._handle_link(target)
                return

    def _flatten_dots(self):
        index = self._path_pos
        while index < len(self._path):
            if self._path[index] == '.':
                # Entries for '.' we can just ignore
                del self._path[index]
                continue

            if self._path[index] == '..':
                if index == 0:
                    raise err.PathError(("Tried to '..' above the root dir " +
                                         "in path '%s'") % self._orig_path)

                # For '..' entries, delete the current component _and_ the
                # previous one
                del self._path[index]
                del self._path[index-1]

                # Go back one step
                index -= 1
                continue

            index += 1

    def _set_vol(self, cell, volname):
        if volname.endswith('.readonly'):
            volname = volname[0:-len('.readonly')]
        elif volname.endswith('.backup'):
            volname = volname[0:-len('.backup')]

        if cell is None:
            cell = self._voldump.cell

        self._voldump = fabs.voldump.find_oldest(cell=cell, rwname=volname,
                                                 after_time=self._timestamp-1)

        if self._voldump is None:
            self._voldump = fabs.voldump.find_latest(cell=cell, rwname=volname,
                                                     before_time=self._timestamp+1)

        if self._voldump is None:
            raise err.PathError(("Path '%s' element %d refers to volume %s, " +
                                 "which appears to not be backed up") % (
                                 self.path_str(), self._path_pos, volname))

    def _handle_link(self, target):
        if (target.startswith('#') or target.startswith('#')) and target.endswith('.'):
            self._handle_mntpt(target)
            return

        self._n_symlinks += 1
        if self._n_symlinks > self.MAX_SYMLINKS:
            raise err.PathError("Path '%s' contains too many (%d) symlinks" % (
                                self._orig_path, self._n_symlinks))

        new_path = self._split_path(target)

        if target.startswith('/'):
            # Replace all path components up to the current one with the new
            # path
            self._path[0:self._path_pos] = new_path

            # Start processing from the root when we continue processing our path
            self._path_pos = 0
            return

        # Replace our current path component with the new target path
        self._path[self._path_pos-1:self._path_pos] = new_path

        # Continue processing our path starting from the root. It's maybe
        # possible to just continue from our current location in the path, but
        # it requires more work (we need to re-evaluate our relative path
        # inside the volume root, etc)
        self._path_pos = 0

    def _handle_mntpt(self, target):
        # Trim off leading '%'/'#', and trailing '.'
        target = target[1:-1]
        parts = target.split(':', 1)

        if len(parts) == 1:
            cell = None
            volname = parts[0]
        else:
            assert len(parts) == 2
            cell = parts[0]
            volname = parts[1]

        self._set_vol(cell, volname)



#
# Copyright (c) 2015-2022, Sine Nomine Associates ("SNA")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND SNA DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL SNA BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

# This file contains some helper and utility functionality to run our tests.
# All tests should probably import this module and run tests through our
# FabsTest class (or subclasses) to get some convenient functionality.

import pytest
import sys
import shlex
import time
import contextlib
import tempfile
import os
import json
import multiprocessing
import copy
import yaml
import subprocess

from unittest import mock
import sqlalchemy as sa
from sqlalchemy.pool import NullPool

import fabs.scripts.fabsys
import fabs.db as db

yield_fixture = pytest.fixture
if int(pytest.__version__.split('.')[0]) < 3:
    # Before pytest 3.0, we need to use a separate @yield_fixture for fixtures
    # that yield. After 3.0, we can use the regular @fixture, and starting in
    # 6.2, using yield_fixture emits a deprecation warning.
    yield_fixture = pytest.yield_fixture

@contextlib.contextmanager
def nullcontext():
    yield

class FabsTest:
    curTest = None
    config = {
        'afs': {
            'cell': 'test.example.com',
        }
    }
    _capfd = None
    _tmp = None
    _out_strs = None
    _err_strs = None

    @yield_fixture(autouse=True)
    def fabs_cap(self, capfd):
        FabsTest.curTest = self
        self._capfd = capfd
        self._out_strs = []
        self._err_strs = []

        yield

        # After the rest has run, print out all of the captured and consumed
        # stdout/stderr data, so it doesn't disappear.
        if self._out_strs:
            sys.stdout.write("\n")
        for ostr in self._out_strs:
            sys.stdout.write(ostr)

        if self._err_strs:
            sys.stderr.write("\n")
        for estr in self._err_strs:
            sys.stderr.write(estr)

        self._out_strs = []
        self._err_strs = []
        self._capfd = None
        FabsTest.curTest = None

    @yield_fixture(autouse=True)
    def fabs_tmp(self, tmpdir):
        self._tmp = tmpdir
        yield
        self._tmp = None

    @yield_fixture
    def tzutc(self):
        prev_tz = os.environ.get('TZ', None)
        os.environ['TZ'] = 'UTC'
        time.tzset()

        yield

        if prev_tz is None:
            del os.environ['TZ']
        else:
            os.environ['TZ'] = prev_tz
        time.tzset()

    def comment(self, astr):
        # .disabled() doesn't exist in pytest 2.9. See:
        # https://github.com/pytest-dev/pytest/issues/1599#issuecomment-261964636
        # for a way to do this directly, if needed
        #with self._capfd.disabled():

        for line in astr.splitlines():
            print('# ' + line, file=sys.stderr)

    def mkftemp(self, prefix, adir=False):
        prefix = '%s.' % prefix
        kwargs = {}

        # Use line-buffered files
        kwargs['buffering'] = 1

        # Use our per-test tmp dir
        kwargs['dir'] = self._tmp

        assert self._tmp is not None

        if adir:
            return tempfile.mkdtemp(prefix=prefix, dir=self._tmp)
        else:
            return tempfile.NamedTemporaryFile(mode='w+', prefix=prefix, delete=False, **kwargs)

    def run_in_fork(self, func, fail=False, init_config=False):
        def target():
            with contextlib.ExitStack() as stack:
                stack.enter_context(db.ensure_reset())
                if init_config:
                    config_path = stack.enter_context(self.config_file())
                    fabs.config.load(cli_conf_file=config_path)
                ret = func()
                if ret is None:
                    ret = 0
            sys.exit(ret)

        child = multiprocessing.Process(target=target)
        child.start()
        child.join()
        assert child.exitcode is not None
        if fail:
            assert child.exitcode != 0
        else:
            assert child.exitcode == 0

    @contextlib.contextmanager
    def config_file(self, skip=False):
        """
        Write out our data in self.config to a tmp file, to use as a FABS
        config file.
        """
        if skip:
            yield None
            return

        config_fh = self.mkftemp('config')
        try:
            if self.config is not None:
                yaml.safe_dump(self.config, stream=config_fh)
                config_fh.flush()
            yield config_fh.name
        finally:
            config_fh.close()

    def _fabsys(self, argstr, fail=False):
        """
        Run the equivalent of 'fabsys <argstr>'.

        So, self._fabsys('foo --bar') is like running 'fabsys foo --bar'.

        Args:
            argstr (str): Arguments to 'fabsys'.
            fail (bool, optional): If True, assert that the command fails
                (exits with nonzero). Otherwise, assert that the command
                succeeds (exits with 0). Defaults to False.
        """

        argv = shlex.split(argstr)

        # Unless --config was explicitly specified in the args, provide a
        # config file to use, so we don't use the default /etc/fabs config file
        # if the machine we're running on has one.
        skip_config=False
        if '--config' in argv:
            skip_config=True

        with self.config_file(skip=skip_config) as config_path:
            if config_path is not None:
                argv.extend(['--config', config_path])

            self.run_in_fork(lambda: fabs.scripts.fabsys.main(argv), fail=fail)

    def _readout(self, stderr, show=False):
        """
        Args:
            stderr (bool): If True, return captured stderr data. Otherwise,
                return captured stdout data.
            show (bool, optional): If True, still print the captured data back
                out afterwards, in addition to returning it here.

        Returns:
            str: Captured data.
        """
        out, err = self._capfd.readouterr()

        if stderr:
            # Return stderr data
            retval = err
            if not show:
                # Don't save the captured stderr
                err = ''

        else:
            # Return stdout data
            retval = out
            if not show:
                # Don't save the captured stdout
                out = ''

        if out:
            self._out_strs.append(out)
        if err:
            self._err_strs.append(err)

        return retval

    def _fabsys_cap(self, argstr, stderr=False, fail=False):
        """
        Same as `fabsys`, but capture the stdout (or stderr) output from the
        fabsys call and return it.
        """
        self._readout(stderr=stderr, show=True)
        self._fabsys(argstr, fail=fail)
        return self._readout(stderr=stderr)

    def fabsys_stdout(self, argstr):
        """
        Run the given fabsys command, capture stdout from it, and return the
        captured stdout string
        """
        return self._fabsys_cap(argstr)

    def fabsys_stderr(self, argstr, fail=True):
        """
        Same as fabsys_stdout(), but capture stderr instead of stdout
        """
        return self._fabsys_cap(argstr, stderr=True, fail=fail)

    def fabsys_json(self, argstr):
        """
        Same as fabsys_stdout(), but run the command with json formatting, and
        return the parsed json object
        """
        return json.loads(self.fabsys_stdout(argstr + ' --format json'))

    def fabsys_comment(self, argstr):
        """
        Same as fabsys_stdout(), but output the captured stdout as a comment
        """
        self.comment(self.fabsys_stdout(argstr))

def curTest():
    """
    Get the object for the current-running test. We can't always easily pass in
    the test case object to some objects, callback functions, etc, so just
    store it globally for convenience.
    """
    assert FabsTest.curTest is not None
    return FabsTest.curTest

# Pick a path that's very unlikely to exist. We'll run all of our commands
# through this, to try to guarantee that an error occurs if something goes
# wrong and the relevant command is actually exec'd.
CMD_WRAPPER='/dev/nonexistent/path'
KEYTAB='/opt/fabs/etc/test.keytab'

class MockCommand:
    """
    This effectively mocks the subprocess.Popen class, using our mocked cell
    data to provide command results and output. We don't implement everything
    that Popen can do, just the functionality that gets used.

    We effectively pretend that the process has exited successfully as soon as
    this object is created, and stdout/stderr are stored in temporary files
    that wil be read and polled in the usual way.

    Just like with a normal subprocess.Popen, you can use these mock command
    handles like this:

        with MockWhatever() as ph:
            pass

    And on exiting the 'with' block, the relevant file handles will be closed.
    """

    def __init__(self, argv, env, data):
        self._argv = argv
        self._env = env
        self._mockdata = data

        test = curTest()

        self.stdin = test.mkftemp('stdin-%s' % argv[0])
        self.stdout = test.mkftemp('stdout-%s' % argv[0])
        self.stderr = test.mkftemp('stderr-%s' % argv[0])

        self.returncode = None

        # Some callers will grab the pid to log; just give an obviously fake
        # pid
        self.pid = 99999999

    def get_vol(self, name):
        return self._mockdata.vols.get(name, None)

    def start(self):
        self.returncode = self._populate_output(self.stdout, self.stderr)

        self.stdout.flush()
        self.stderr.flush()

        self.stdout.seek(0)
        self.stderr.seek(0)

    def _cleanup(self):
        self.stdin.close()
        self.stdout.close()
        self.stderr.close()

    def __enter__(self):
        return self
    def __exit__(self, *exc_info):
        self._cleanup()

    def _populate_output(self, stdout, stderr):
        raise NotImplementedError

    # pylint: disable=redefined-builtin
    def communicate(self, input=None):
        if self.returncode == 0:
            # Seek back to the end of our stdout/stderr files
            self.stdout.seek(0, 2)
            self.stderr.seek(0, 2)

            self.returncode = self._do_communicate(input, self.stdout, self.stderr)

            self.stdout.flush()
            self.stderr.flush()

            self.stdout.seek(0)
            self.stderr.seek(0)

        ret = (self.stdout.read().encode('utf-8'),
               self.stderr.read().encode('utf-8'))
        self._cleanup()
        return ret

    def poll(self):
        assert self.returncode is not None
        self._cleanup()
        return self.returncode

    def wait(self):
        return self.poll()

    def _do_communicate(self, stdin_bytes, stdout, stderr):
        return 0

    @classmethod
    def popen(cls, argv, env, data):
        return cls(argv, env, data)

class _NO_DEFAULT:
    pass

# pylint: disable=abstract-method
class MockAFSCommand(MockCommand):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._args = self.cli_to_args(self._argv)

    def vlnoent(self, stderr):
        stderr.write("VLDB: no such entry\n")
        return 255

    # Transforms e.g. '-foo bar -baz' into {'foo': 'bar', 'baz': True}
    @classmethod
    def cli_to_args(cls, argv):
        ret = {}
        key = None
        
        # Ignore the first two arguments (e.g. 'vos examine')
        for arg in argv[2:]:
            if arg.startswith('-'):
                key = arg[1:]
                ret[key] = True
            else:
                assert key is not None
                ret[key] = arg
                key = None
        return ret

    def arg(self, name, default=_NO_DEFAULT):
        if default is _NO_DEFAULT:
            assert name in self._args
        elif name not in self._args:
            return default
        return self._args[name]

class MockDumpFilter(MockCommand):
    def _populate_output(self, stdout, stderr):
        stdout.write("{}")
        return 0

class MockK5start:
    @classmethod
    def popen(cls, argv, env, data):
        argv = argv[:]
        # Remove the beginning of 'argv' until we hit the '--' arg. After the
        # '--' arg, our wrapped command follows.
        while argv.pop(0) != '--':
            pass
        return mock_popen(argv, env, data)

class MockVosListaddrs(MockAFSCommand):
    def _populate_output(self, stdout, stderr):
        assert self.arg('noresolv') is True
        assert self.arg('printuuid') is True

        for fs in self._mockdata.fs_list:
            stdout.write("UUID: %s\n" % fs.uuid)
            stdout.write("%s\n" % fs.ip)
            stdout.write("\n")
        return 0

class MockVosBackup(MockAFSCommand):
    def _populate_output(self, stdout, stderr):
        volid = self.arg('id')
        vol = self.get_vol(volid)
        if vol is None:
            return self.vlnoent(stderr)

        if not vol.online:
            stderr.write("\n")
            stderr.write("Could not re-clone backup volume %d\n" % vol.rwid)
            stderr.write("    Volume is off line\n")
            stderr.write("Volume is off line\n")
            stderr.write("Error in vos backup command.\n")
            stderr.write("Volume is off line\n")
            return 255

        stdout.write("Created backup volume for %s\n" % volid)
        return 0

class MockVosExamine(MockAFSCommand):
    def _examine(self, vol, stdout):
        data = {
            'name': vol.name,
            'volid': vol.rwid,
            'typestr': 'RW',
            'size': vol.size/1024,
            'online': 'On-line',
            'need_salvage': '',
            'fs': vol.fs.ip,
            'part': vol.part,
            'rwid': vol.rwid,
            'roid': vol.roid,
            'bkid': vol.bkid,
            'clid': vol.clid,
            'maxquota': vol.maxquota,
            'crdate': vol.crdate,
            'cpdate': vol.cpdate,
            'bkdate': vol.bkdate,
            'ladate': vol.ladate,
            'ludate': vol.ludate,
            'n_access': vol.n_access,
        }

        if not vol.online:
            data['online'] = 'Off-line'

        stdout.write("""%(name)-32s %(volid)10lu %(typestr)s %(size)d K  %(online)s%(need_salvage)s
    %(fs)s %(part)s 
    RWrite %(rwid)10lu ROnly %(roid)10lu Backup %(bkid)10lu 
    MaxQuota %(maxquota)10d K 
    Creation    %(crdate)s
    Copy        %(cpdate)s
    Backup      %(bkdate)s
    Last Access %(ladate)s
    Last Update %(ludate)s
    %(n_access)d accesses in the past day (i.e., vnode references)
""" % data)

        return self._vldata(vol, stdout)

    def _listvldb_one(self, vol, stdout):
        stdout.write("\n%s\n" % vol.name)
        return self._vldata(vol, stdout)

    def _listvldb_all(self, stdout):
        stdout.write("VLDB entries for all servers\n")
        vol_list = self._mockdata.vol_list
        for vol in vol_list:
            stdout.write("\n%s\n" % vol.name)
            code = self._vldata(vol, stdout)
            if code != 0:
                return code
        stdout.write("\nTotal entries: %d\n" % len(vol_list))
        return 0

    def _vldata(self, vol, stdout):
        data = {
            'rwid': vol.rwid,
            'roid': vol.roid,
            'bkid': vol.bkid,
            'clid': vol.clid,
            'fs': vol.fs.ip,
            'part': vol.part,
            'site_type': 'RW',
            'flag_str': '',
            'n_sites': 1,
        }

        stdout.write("""    RWrite: %(rwid)-10u    ROnly: %(roid)-10u    Backup: %(bkid)-10u    RClone: %(clid)-10lu
    number of sites -> %(n_sites)lu
""" % data)

        stdout.write("        server %(fs)s partition %(part)s %(site_type)s Site %(flag_str)s\n" % data)
        return 0

    def _populate_output(self, stdout, stderr):
        cmd = self._argv[1]
        if cmd == 'listvldb':
            volid = self.arg('name', default=None)
            if volid is None:
                return self._listvldb_all(stdout)
        else:
            assert cmd == 'examine'
            volid = self.arg('id')

        vol = self.get_vol(volid)
        if vol is None:
            return self.vlnoent(stderr)
        elif cmd == 'listvldb':
            return self._listvldb_one(vol, stdout)
        else:
            return self._examine(vol, stdout)

class MockVosSize(MockAFSCommand):
    def _populate_output(self, stdout, stderr):
        assert self.arg('dump') is True

        volid = self.arg('id')
        vol = self.get_vol(volid)

        if vol is None:
            return self.vlnoent(stderr)

        stdout.write("Volume: %s\ndump_size: %d\n" % (volid, vol.size))
        return 0

class MockVosDump(MockAFSCommand):
    def _populate_output(self, stdout, stderr):
        volid = self.arg('id')
        vol = self.get_vol(volid)
        if vol is None:
            return self.vlnoent(stderr)

        stdout.write("mock vos dump data:%s" % vol.rwid)
        return 0

class MockDumpReport(MockCommand):
    def _populate_output(self, stdout, stderr):
        return 0

    def _do_communicate(self, stdin_bytes, stdout, stderr):
        self._mockdata.report(stdin_bytes.decode('utf-8'))
        return 0

class MockDumpscan(MockCommand):
    def _populate_output(self, stdout, stderr):
        flags = self._argv[1]

        if flags == '-Pvip':
            # Print out vnode info; just output nothing for now
            return 0

        assert self._argv[1] == "-PV"
        # -PV means just print out some volume metadata

        fname = self._argv[2]
        with open(fname, 'r') as fh:
            volname = fh.read().split(':')[1]

        vol = self.get_vol(volname)
        assert vol is not None

        data = {
            'volid': vol.rwid,
            'volname': volname,
            'rwid': vol.rwid,
            'clid': vol.clid,
            'maxquota': vol.maxquota,
            'size': vol.size,
            'crdate': vol.crdate,
            'ladate': vol.ladate,
            'ludate': vol.ludate,
            'bkdate': vol.bkdate,
            'n_access': vol.n_access,
        }

        stdout.write("""* VOLUME HEADER [36 = 0x0000000000000024]
 Volume ID:   %(volid)d
 Version:     1
 Volume name: %(volname)s
 In service?  true
 Blessed?     true
 Uniquifier:  5
 Type:        0
 Parent ID:   %(rwid)d
 Clone ID:    %(clid)d
 Max quota:   %(maxquota)d
 Min quota:   0
 Disk used:   %(size)d
 File count:  5
 Account:     0
 Owner:       0
 Created:     %(crdate)s
 Accessed:    %(ladate)s
 Updated:     %(ludate)s
 Expires:     Wed Dec 31 18:00:00 1969
 Backed up:   %(bkdate)s
 Offine Msg:  volume has been soft detached
 MOTD:        
 Weekuse:              0          0          0          0
 Weekuse:              0          0          0
 Dayuse Date: Thu Aug 10 00:00:00 2017
 Daily usage: %(n_access)d
""" % data)

        return 0

_mocked_commands = {
    ('k5start',): MockK5start,
    ('vos', 'backup'): MockVosBackup,
    ('vos', 'dump'): MockVosDump,
    ('vos', 'examine'): MockVosExamine,
    ('vos', 'listvldb'): MockVosExamine,
    ('vos', 'listaddrs'): MockVosListaddrs,
    ('vos', 'size'): MockVosSize,
    ('dump-filter',): MockDumpFilter,
    ('dumpscan',): MockDumpscan,
    ('dump-report',): MockDumpReport,
}

def mock_popen(argv, env, data, **kwargs):
    wrapper = argv[0]
    assert wrapper == CMD_WRAPPER

    argv = argv[1:]
    cmd = argv[:]
    while cmd:
        t_cmd = tuple(cmd)
        if t_cmd in _mocked_commands:
            proc = _mocked_commands[t_cmd].popen(argv, env, data)
            proc.start()
            return proc
        cmd.pop()

    raise Exception("cmd %r not found in _mocked_commands" % argv)

class MockVol:
    def __init__(self, mockdata, fs, part, name=None):
        self.mockdata = mockdata

        if name is None:
            name = self.get_name()
        self.name = name

        self.fs = fs
        self.part = part

        self.size = 4096
        self.online = True

        self.rwid = self.get_id()
        self.roid = self.get_id()
        self.bkid = self.get_id()
        self.clid = self.get_id()

        self.crdate = 'Mon Jul  6 19:13:11 2015'
        self.cpdate = 'Mon Jul  6 19:13:11 2015'
        self.bkdate = 'Mon Jul  6 19:13:11 2015'
        self.ladate = 'Mon Jul  6 19:13:11 2015'
        self.ludate = 'Mon Jul  6 19:13:11 2015'

        self.n_access = 5

        self.maxquota = 5000

    def ids(self):
        return [self.rwid, self.roid, self.bkid, self.clid]

    def set_online(self, online=True):
        self.online = online

    def get_id(self):
        ret = self.mockdata.volid_counter
        self.mockdata.volid_counter += 1
        return ret

    def get_name(self):
        ret = 'vol.mock.%d' % self.mockdata.volname_counter
        self.mockdata.volname_counter += 1
        return ret

class MockFS:
    def __init__(self, mockdata):
        self.mockdata = mockdata
        self.ip = self.get_ip()
        self.uuid = self.get_uuid()

    def get_ip(self):
        ret = '127.0.0.%d' % self.mockdata.fsip_counter
        self.mockdata.fsip_counter += 1
        return ret

    def get_uuid(self):
        ret = '%08x-%04x-%04x-%02x-%02x-%012x' % (
              1, 2, 3, 4, 5, self.mockdata.fsuuid_counter)
        self.mockdata.fsuuid_counter += 1
        return ret

class MockCellData:
    def __init__(self, report_fh):
        self.volname_counter = 1
        # 0x20000000, the number that the vlserver starts counting volids from
        self.volid_counter = 536870912

        self.fsip_counter = 1
        self.fsuuid_counter = 1

        self._report_fh = report_fh

        fs = MockFS(self)
        self.fs_list = [fs]
        self.vols = {}

        vol_list = [MockVol(self, fs, '/vicepa', 'root.cell')]
        for _ in range(10):
            vol_list.append(MockVol(self, fs, '/vicepa'))

        self.vol_list = vol_list
        for vol in vol_list:
            self.vols[vol.name] = vol
            for volid in vol.ids():
                self.vols[str(volid)] = vol

    def report(self, data):
        self._report_fh.write(data)
        self._report_fh.flush()

    def report_data(self):
        self._report_fh.seek(0)
        return self._report_fh.read()

class FabsMockEnvTest(FabsTest):
    mockdata = None
    db_data = None

    config_base = {
## Uncomment to show a ton of debug info while test backups are running
#        'log': {
#            'level': 'debug',
#        },
        'afs': {
            'cell': 'test.example.com',
            'dumpscan': [CMD_WRAPPER, 'dumpscan'],
            'vos': [CMD_WRAPPER, 'vos'],
            'fs': [CMD_WRAPPER, 'fs'],
            'pts': [CMD_WRAPPER, 'pts'],
            'aklog': [CMD_WRAPPER, 'aklog'],
            'keytab': KEYTAB,
        },
        'k5start': {
            'command': [CMD_WRAPPER, 'k5start'],
        },
        'db': {
            'url': None,
            'batch_size': 3,
        },
        'dump': {
            'include': {
                'volumes': ['*'],
            },
            'storage_dirs': None,
            'filter_cmd': {
                'json': [CMD_WRAPPER, 'dump-filter'],
            },
        },
        'stage': {
            'notify_cmd': [CMD_WRAPPER, 'stage-notify'],
        },
        'backend': {
            'request_cmd': [CMD_WRAPPER, 'backend-request'],
        },
        'report': {
            'json': {
                'command': [CMD_WRAPPER, 'dump-report'],
            },
            'txt': {
                'command': None,
            },
        },
        'lockdir': None,
    }

    @yield_fixture(autouse=True)
    def fabs_mock_data(self, fabs_tmp, fabs_cap):
        self.config = copy.deepcopy(self.config_base)

        self.config['dump']['storage_dirs'] = [self.mkftemp('storage', adir=True)]
        self.config['lockdir'] = self.mkftemp('lockdir', adir=True)
        with self.mkftemp('db') as db_fh:
            self.config['db']['url'] = 'sqlite:///%s' % db_fh.name

        with self.mkftemp('report') as report_fh:
            mockdata = MockCellData(report_fh)
            with self._patch_popen(mockdata):
                self.mockdata = mockdata
                yield mockdata
                self.mockdata = None

    @pytest.fixture(autouse=True)
    def db_url(self, raw_db_url, fabs_mock_data):
        if raw_db_url is not None:
            # Drop all tables in given db
            def reset_db():
                engine = sa.create_engine(raw_db_url, poolclass=NullPool)
                with engine.connect() as conn:
                    meta = sa.MetaData(engine)
                    meta.reflect(bind=conn)
                    meta.drop_all()
                return 0
            self.run_in_fork(reset_db)

            # Point our FABS config to the given db
            self.config['db']['url'] = raw_db_url

        return self.config['db']['url']

    @contextlib.contextmanager
    def _patch_popen(self, data):
        _orig_popen = subprocess.Popen

        def new_popen(argv, env=None, **kwargs):
            if isinstance(argv, list) and argv[0] == CMD_WRAPPER:
                return mock_popen(argv, env, data)
            return _orig_popen(argv, env=env, **kwargs)

        with mock.patch('subprocess.Popen', new=new_popen):
            # Silence warnings about our aklog command not being runnable
            # (since we mock actual commands, our aklog points to a bogus path)
            with mock.patch('fabs.util._check_aklog.warned', new=True):
                yield

    def _db_connargs(self, url):
        if url.drivername == 'sqlite':
            return [url.database]

        if url.drivername == 'mysql':
            args = []
            if url.username:
                args.append('--user=%s' % url.username)
            if url.password:
                args.append('--password=%s' % url.password)
            if url.host:
                args.append('--host=%s' % url.host)
            if url.port:
                args.append('--port=%s' % url.port)

            assert url.database
            args.append(url.database)

            return args

        raise ValueError("Unhandled db dialect %r" % url.drivername)

    def db_bulk_dump(self, db_url):
        url = sa.engine.url.make_url(db_url)

        if url.drivername == 'sqlite':
            args = ['sqlite3'] + self._db_connargs(url) + ['.dump']
        elif url.drivername == 'mysql':
            args = ['mysqldump', '--compact'] + self._db_connargs(url)
        else:
            raise ValueError("Unhandled db dialect %r" % url.drivername)

        res = subprocess.run(args, check=True, stdout=subprocess.PIPE)
        return res.stdout.decode('ascii')

    def db_bulk_load(self, a_url, a_sql):
        url = sa.engine.url.make_url(a_url)

        sql = ''.join([
            'BEGIN;',
            a_sql,
            'COMMIT;'
        ]).encode('utf-8')

        if url.drivername == 'sqlite':
            args = ['sqlite3'] + self._db_connargs(url)

        elif url.drivername == 'mysql':
            args = ['mysql'] + self._db_connargs(url)
        else:
            raise ValueError("Unhandled db dialect %r" % url.drivername)

        subprocess.run(args, input=sql, check=True)

    def do_backup_init(self):
        self.fabsys_comment('db-init --exec')
        if self.db_data is not None:
            self.db_bulk_load(self.config['db']['url'], self.db_data)
        self.fabsys_comment('storage-init --all')

    def do_backup_run(self, volname=None):
        self.comment("Starting backup")
        if volname is None:
            res = self.fabsys_json('backup-start --all')
        else:
            res = self.fabsys_json('backup-start --volume %s' % volname)

        assert res['fabs_backup_start']['brun_id'] is not None
        brid = res['fabs_backup_start']['brun_id']

        self.comment("Running backup")
        n_steps = 30
        for _ in range(n_steps):
            self.fabsys_comment('server --once')

        return brid

    def do_backup_success(self, volname=None):
        brid = self.do_backup_run(volname)

        # Check that the backup run finished
        res = self.fabsys_json('backup-status %d' % brid)

        brinfo = res['fabs_backup_status']['status'][0]
        assert brinfo['id'] == brid
        assert brinfo['state'] == 'DONE'

        res = json.loads(self.mockdata.report_data())

        assert res['success']['id'] == brid

        return res

# Kinda hack-y, but this turns off some consistency options in sqlite. We don't
# care about on-disk consistency just for the tests, and this makes the tests
# run a bit faster.
@sa.event.listens_for(sa.engine.Engine, 'connect')
def db_pragma(dbapi_conn, conn_record):
    if type(dbapi_conn).__module__ == 'sqlite3':
        # Only do this for sqlite3 conns. This check is a bit iffy, but there
        # doesn't seem to be a better way to check for the conn's dialect.
        cursor = dbapi_conn.cursor()
        cursor.execute("PRAGMA synchronous = OFF")
        cursor.execute("PRAGMA journal_mode = MEMORY")
        cursor.close()



#
# Copyright (c) 2015-2022, Sine Nomine Associates ("SNA")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND SNA DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL SNA BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

# This module just stores some fabs-related constants. You can access them
# like fabs.const.VERSION, or in a dict like fabs.const.consts['VERSION'].
# The latter is just for convenience of looping through all available
# constants.

consts = {'PREFIX': '/usr', 'SYSCONFDIR': '/etc', 'LOCALSTATEDIR': '/var', 'LOCKDIR': '/var/lock', 'VERSION': '1.1', 'CONF_DIR': '/etc/fabs', 'STORAGE_DIR': '/var/lib/fabs'}

# Set fabs.const.VERSION to consts['VERSION'], etc
globals().update(consts)



#
# Copyright (c) 2021-2022, Sine Nomine Associates ("SNA")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND SNA DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL SNA BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

import json
import pytest
import select

import fabs.testutil as tu
import fabs.err

import fabs.util as util

# Tests for internal util functions.

class TestUtil(tu.FabsTest):
    def test_timeout(self):
        select_finished = False

        # Set a timeout of 1 second, and sleep (via select()) for 2 seconds.
        # The sleep should never finish, since the timeout should trigger well
        # before then.
        with util.syscall_timeout(1):
            select.select([], [], [], 2)
            select_finished = True

        assert not select_finished

    def test_timeout_nest(self):
        with pytest.raises(fabs.err.InternalError):
            with util.syscall_timeout(5):
                with util.syscall_timeout(5):
                    pass

    def test_time_str2unix(self):
        assert 0 == util.time_str2unix('@0')
        assert 1644102842 == util.time_str2unix('2022-02-05 23:14:02 UTC')
        assert 1644102842 == util.time_str2unix('@1644102842')
        assert 2117488442 == util.time_str2unix('2037-02-05 23:14:02 UTC')

    def test_time_unix2str(self, tzutc):
        assert 'Sat Feb 05 23:14:02 2022 UTC' == util.time_unix2str(1644102842)
        assert 'Thu Feb 05 23:14:02 2037 UTC' == util.time_unix2str(2117488442)

    @pytest.mark.parametrize('data', [
                                      [],
                                      [1],
                                      [1,2,3,4,5],
                                     ])
    def test_json_generator(self, data, tmpdir):
        def genfunc():
            for item in data:
                yield item

        obj = {'key': util.json_generator(genfunc())}
        res = json.dumps(obj)
        new_obj = json.loads(res)

        assert {'key': data} == new_obj

        obj = {'key': util.json_generator(genfunc())}
        with open(tmpdir + '/foo', 'w') as fh:
            json.dump(obj, fh)
        with open(tmpdir + '/foo') as fh:
            self.comment(fh.read())
        with open(tmpdir + '/foo') as fh:
            new_obj = json.load(fh)

        assert {'key': data} == new_obj

    @pytest.mark.parametrize('data,exp',
                             [
                              (['foo'],
                               'foo'),
                              (['foo','bar'],
                               'foo and bar'),
                              (['foo','bar','baz'],
                               'foo, bar, and baz'),
                              (['foo','bar','baz','quux'],
                               'foo, bar, baz, and quux'),
                             ])
    def test_list2str(self, data, exp):
        assert exp == util.list2str(data)



#
# Copyright (c) 2015-2022, Sine Nomine Associates ("SNA")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND SNA DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL SNA BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

import json

from fabs.dumpjob import DumpJob

import fabs.server
import fabs.util as util
import fabs.err as err
import fabs.config as config
import fabs.const
import fabs.log
log = fabs.log.getLogger(__name__)

def report_success(brun):
    """
    Run the relevant report-generating commands for a successful backup run.

    Args:
        `fabs.brun.BackupRun`: The backup run that successfully ran.
    """

    jobs = DumpJob.describe(brun)
    job_data = {}
    jobs_done = 0
    jobs_err = 0
    jobs_skipped = 0
    for job in jobs:
        if job['state'] == 'DONE':
            job['failed'] = False
            if job['skip_reason']:
                jobs_skipped = jobs_skipped + 1
            else:
                jobs_done = jobs_done + 1
        else:
            job['failed'] = True
            jobs_err = jobs_err + 1

        srv = job_data.setdefault(job['server_addr'], {})
        part = srv.setdefault(job['partition'], [])
        part.append(job)

    data = dict(
        fabs_version=fabs.const.VERSION,
        id=brun.id,
        note=brun.note,
        cell=brun.cell,
        volume=brun.volume,
        nvols_total=len(jobs),
        nvols_done=jobs_done,
        nvols_err=jobs_err,
        nvols_skipped=jobs_skipped,
        jobs=job_data,
    )

    for attr in ('start', 'end'):
        val = getattr(brun, attr)
        log.d("Got attribute %s value %s" % (attr, val))
        data[attr] = util.time_unix2str(val)

    if data['volume'] is None:
        data['volume'] = '*'

    report_data = dict(
        success=data,
    )

    if config.get('report/only_on_error') and data['nvols_err'] == 0:
        log.d("Not sending success-only report")
        return

    cmd = config.get('report/txt/command')
    if cmd:
        report = _report_txt(report_data)
        util.run_config_cmd(cmd, stdin=report)

    cmd = config.get('report/json/command')
    if cmd:
        report = json.dumps(report_data)
        util.run_config_cmd(cmd, stdin=report)

def _report_txt(data):
    report = ""

    if 'success' in data:
        data = data['success']
        report += """FABS v%(fabs_version)s Backup Report

Backup run %(id)s for cell %(cell)s, volume %(volume)s
Started on:  %(start)s
Finished on: %(end)s
Note: %(note)s

Attempted to backup %(nvols_total)d volume(s)
Successfully dumped %(nvols_done)d volume(s)
Failed to dump %(nvols_err)d volume(s)
Skipped %(nvols_skipped)d volume(s)

""" % data

        report += "Volumes per partition:\n"

        failed_jobs = []
        done_jobs = []
        error_jobs = []
        skipped_jobs = {}

        for server, part_data in data['jobs'].items():
            for part, joblist in part_data.items():
                done = 0
                failed = 0
                skipped = 0
                errors = 0
                for job in joblist:
                    if job['failed']:
                        failed = failed + 1
                        failed_jobs.append(job)

                    elif job['skip_reason']:
                        skipped += 1
                        skipped_jobs.setdefault(job['skip_reason'], [])
                        skipped_jobs[job['skip_reason']].append(job)

                    else:
                        if job['errors'] > 0:
                            errors = errors + 1
                            error_jobs.append(job)

                        done = done + 1
                        done_jobs.append(job)

                error_str = ""
                if errors > 0:
                    error_str = " (%d with errors)" % errors

                report += "  %s vicep%s: %d total, %d dumped%s, %d failed, %d skipped\n" % (
                          server, part, len(joblist), done, error_str, failed, skipped)

        if failed_jobs:
            report += "\nVolumes that failed to dump:\n"

            # Sort jobs by volume name
            failed_jobs = sorted(failed_jobs, key=lambda job: job['name'])
            for job in failed_jobs:
                report += "  %s (vl_id %d)\n" % (job['name'], job['vl_id'])

        if error_jobs:
            report += "\nThe following volumes encountered non-fatal errors.\n"
            report += "They were dumped, but encountered some problems along\n"
            report += "the way. Look for log messages mentioning the relevant\n"
            report += "volume name or FABS vl_id for more details.\n"
            report += "\n"
            report += "Volumes dumped with errors:\n"

            # Sort jobs by volume name
            error_jobs = sorted(error_jobs, key=lambda job: job['name'])
            for job in error_jobs:
                report += "  %s (vl_id %d, errors %d)\n" % (job['name'],
                          job['vl_id'], job['errors'])

        for skip_reason in sorted(skipped_jobs.keys()):
            jobs = sorted(skipped_jobs[skip_reason], key=lambda job: job['name'])
            if skip_reason == 'UNCHANGED':
                skip_txt = 'because the volume was unchanged'
            elif skip_reason == 'OFFLINE':
                skip_txt = 'because the volume was offline'
            else:
                skip_txt = 'for unrecognized reason "%s"' % skip_reason

            report += "\nVolumes that were skipped %s:\n" % skip_txt

            for job in jobs:
                report += "  %s (vl_id %d)\n" % (job['name'], job['vl_id'])

        if done_jobs:
            report += "\nVolumes that were dumped successfully:\n"

            done_jobs = sorted(done_jobs, key=lambda job: job['name'])
            for job in done_jobs:
                report += "  %s (vl_id %d)\n" % (job['name'], job['vl_id'])

        return report

    if 'failed' in data:
        raise NotImplementedError

    raise err.InternalError("Cannot format report with keys %r" % data.keys())



"""
FABS
-----

FABS (the "Flexible AFS Backup System") is a suite of tools used for backing
up AFS cells, to be used together with separate (possibly existing) "plain"
backup system.
"""

import glob
import json
import os.path
import pathlib
import subprocess
import sys

from setuptools import setup
from setuptools.command.build_py import build_py

with open('.version') as fh:
    __version__ = fh.readline().rstrip().replace('~', '')

# Set our "build-time" constants, which will be available in fabs.const when
# fabs is built.
_const_dict = {}
_const_dict['PREFIX'] = os.environ.get('PREFIX', '/opt/fabs')
_const_dict['SYSCONFDIR'] = os.environ.get('SYSCONFDIR',
                                           os.path.join(_const_dict['PREFIX'],
                                                        'etc'))
_const_dict['LOCALSTATEDIR'] = os.environ.get('LOCALSTATEDIR',
                                              os.path.join(_const_dict['PREFIX'],
                                                           'var'))
_const_dict['LOCKDIR'] = os.environ.get('LOCKDIR',
                                        os.path.join(_const_dict['LOCALSTATEDIR'],
                                                     'lock'))
_const_dict['VERSION'] = __version__
_const_dict['CONF_DIR'] = os.path.join(_const_dict['SYSCONFDIR'], 'fabs')
_const_dict['STORAGE_DIR'] = os.path.join(_const_dict['LOCALSTATEDIR'],
                                          'lib', 'fabs')

def _consts_code():
    code = [
        'consts = %r' % _const_dict,
        '',
        "# Set fabs.const.VERSION to consts['VERSION'], etc",
        'globals().update(consts)'
    ]
    return ''.join(line+'\n' for line in code)

def _gen_const_py():
    from distutils import log

    # Path to top-level 'fabs' dir.
    topdir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'fabs')

    # Generate const.py by appending our const dict to const.py.in. Write to
    # const.py.tmp, and rename() it to const.py when we're done.

    const_py = os.path.join(topdir, 'const.py')
    const_py_in = os.path.join(topdir, 'const.py.in')
    const_py_tmp = os.path.join(topdir, 'const.py.tmp')
    log.info("generating %s" % const_py)

    with open(const_py_in) as in_fh:
        in_data = in_fh.read()

    with open(const_py_tmp, 'w') as tmp_fh:
        tmp_fh.write(in_data)
        tmp_fh.write(_consts_code())

    os.rename(const_py_tmp, const_py)

def _run(argv):
    arg_str = ' '.join(argv)
    print("+ %s" % arg_str, file=sys.stderr)
    subprocess.run(argv, check=True)

# Subclass the 'build_py' command to run a few extra things at build time.
class fabs_build_py(build_py):
    def run(self):
        # Generate fabs/const.py
        _gen_const_py()

        # Generate manpages in doc/man1
        _run(['./doc/generate-man', 'doc'])

        super().run()

def get_data_files():
    man_paths = []

    for path in glob.iglob('doc/pod1/*.pod'):
        fname = os.path.basename(path)
        assert fname.endswith('.pod')

        page_name = fname[:-len('.pod')]

        man_paths.append('doc/man1/%s.1' % page_name)

    yield ('share/man/man1', man_paths)

topdir = pathlib.Path(__file__).parent
readme_path = topdir / "README.md"
readme_contents = readme_path.read_text()

setup(
    name='fabs',
    version=__version__,
    license='ISC',
    author='Sine Nomine Associates',
    author_email='sna-packager@sinenomine.net',
    url='https://github.com/openafs-contrib/fabs',
    description='A flexible backup suite for AFS cells',
    long_description=readme_contents,
    long_description_content_type='text/markdown',
    install_requires=[
        'PyYAML>=3.10',
        'SQLAlchemy>=0.8',
        'python_dateutil',
    ],
    extras_require={
        'Alembic': ['Alembic>=0.4'],
    },
    cmdclass={
        'build_py': fabs_build_py,
    },
    packages=['fabs', 'fabs.scripts'],
    entry_points={
        # Entry points for our CLI commands
        'console_scripts': [
            'fabsys = fabs.scripts.fabsys:main',
            'sbaf = fabs.scripts.sbaf:main',
        ],
    },
    # Include non-*.py files in our 'fabs' subdir (e.g. log_cli.conf)
    package_data={'fabs': [
        'log_cli.conf',
        'log_daemon.conf',
    ]},

    # Include files outside of our 'fabs' subdir (man pages, hook examples,
    # etc)
    data_files=list(get_data_files()),
)



#
# Copyright (c) 2015-2022, Sine Nomine Associates ("SNA")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND SNA DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL SNA BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

import dateutil.parser
import itertools
from multiprocessing import Process
import os
import os.path
import time
import hashlib
import subprocess
import traceback
import contextlib
import signal

import fabs.err as err
import fabs.db as db
import fabs.config as config
import fabs.log
log = fabs.log.getLogger(__name__)

class FabsProcess(Process):
    """
    Small wrapper around the real Process class.

    In addition to launching the child process with the specified function,
    this does the following:

    In the parent, resets our db connections (they're not supposed to be shared
    across forks). We must do this in the parent, not the child, contrary to
    what the SQLALchemy docs say. If this is done in the child and we're using
    MySQL/MariaDB, the child process will close down its connections on the
    sockets dup'd from the parent, making the parent connections unusable and
    causing "MySQL server has done away" errors to be thrown.

    We also reset our db connections in the child after the target function has
    been run, so they get destroyed cleanly. The normal code for destroying the
    handles usually doesn't get run, since the child process exits with
    `os._exit`, bypassing any 'finally' blocks in our callers or other cleanup
    code.
    """

    # @todo Consider storing all created processes in a weakref dict, and when the
    # server shuts down, it can wait for all spawned children to also exit. Just
    # store them as keys in a WeakKeyDictionary, and use popitem() to get them out.

    # pylint: disable=arguments-differ
    def start(self, *args, **kwargs):
        db.reset()
        super().start(*args, **kwargs)

    def run(self):
        with db.ensure_reset():
            super().run()

def _executable_file(path):
    return os.path.exists(path) and not os.path.isdir(path) and os.access(path, os.X_OK)

def _check_aklog(aklog):
    """
    Check if we have a usable AKLOG for use with k5start. If we don't check
    this and we don't have an aklog, k5start can give a somewhat confusing
    error message.
    """

    if _check_aklog.warned:
        # Only warn about this once
        return None

    abs_aklog = None
    if os.path.isabs(aklog):
        abs_aklog = aklog
    else:
        # Search for aklog in PATH
        for pdir in os.environ['PATH'].split(':'):
            abspath = os.path.join(pdir, aklog)
            if _executable_file(abspath):
                abs_aklog = abspath
                break

    if abs_aklog is None:
        log.warn('aklog_notfound', "Could not find aklog command '%s' in PATH" % aklog)
        _check_aklog.warned = True

    elif not _executable_file(abs_aklog):
        log.warn('aklog_noexec', "aklog command '%s' is not executable" % abs_aklog)
        _check_aklog.warned = True
_check_aklog.warned = False

def _auth_cmd(command, defaults):
    if config.get('afs/localauth'):
        # Run with -localauth if localauth is turned on
        defaults['localauth'] = True
    else:
        # Without localauth, we need to run through k5start
        k5start = list(config.get('k5start/command'))
        aklog = list(config.get('afs/aklog'))
        keytab = config.get('afs/keytab')
        princ = config.get('afs/keytab_princ')
        if princ is None:
            princ = '-U'

        os.environ['AKLOG'] = ' '.join(aklog)
        _check_aklog(aklog[0])

        # Prepend to 'command'
        command[:0] = k5start + ['-q', '-t', '-f', keytab, princ, '--']

def _vos(auth, defaults):
    command = list(config.get('afs/vos'))

    defaults['cell'] = config.get('afs/cell')

    if auth:
        _auth_cmd(command, defaults)
    else:
        defaults['noauth'] = True

    return fabs.cmd.Vos(vos=command, defaults=defaults)

def vos_auth(**defaults):
    return _vos(1, defaults)

def vos_unauth(**defaults):
    return _vos(0, defaults)

def pts_auth():
    command = config.get('afs/pts')
    defaults = {}
    _auth_cmd(command, defaults)
    return fabs.cmd.Pts(pts=command, defaults=defaults)

def fs_auth():
    command = config.get('afs/fs')
    defaults = {}
    _auth_cmd(command, defaults)
    return fabs.cmd.Fs(fs=command, defaults=defaults)

def dumpscan():
    return fabs.cmd.DumpScan(afsdump_scan=config.get('afs/dumpscan'))

class PrettyBytes:
    def __init__(self, total):
        self._last_bytes = 0
        self._last_time = time.time()

        self.bytes = 'unknown bytes'
        self.rate = 'unknown bytes/s'
        self.total = self._pretty(total)

    def _pretty(self, raw):
        for suffix in ('bytes', 'kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB'):
            if raw < 1024:
                return "%.2f %s" % (raw, suffix)
            raw = raw / 1024
        # Getting here is a bit weird, but it's not worth throwing an error for
        return "unknown bytes"

    def update_path(self, path):
        self.update_bytes(os.path.getsize(path))

    def update_bytes(self, cur_bytes):
        cur_time = time.time()

        diff_bytes = cur_bytes - self._last_bytes
        diff_time = cur_time - self._last_time
        raw_rate = diff_bytes / diff_time

        self.bytes = self._pretty(cur_bytes)
        self.rate = "%s/s" % self._pretty(raw_rate)

        self._last_bytes = cur_bytes
        self._last_time = cur_time

def checksum(algo, path):
    """
    Get the checksum of the given file, using the given algorithm.
    """
    m = hashlib.new(algo)

    with open(path, 'rb', buffering=0) as fh:
        while True:
            buf = fh.read(config.get('bufsize'))
            if not buf:
                break
            m.update(buf)

    return m.hexdigest()

def run_config_cmd(cmd, stdin=None, extra_env=None):
    try:
        if isinstance(stdin, str):
            # We need to give our subprocess raw bytes, not a unicode string.
            # Encode the data we give to the subprocess in utf-8.
            stdin = stdin.encode('utf-8')

        env = None
        if extra_env is not None:
            env = os.environ.copy()
            env.update(extra_env)

        with subprocess.Popen(cmd, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE, env=env) as proc:
            log.d("Running configured command %r in pid %d" % (cmd, proc.pid))

            (stdout, stderr) = proc.communicate(stdin)
            if proc.returncode != 0:
                raise err.ConfigCmdError(("Configured command '%s' exited with status " +
                                          "%d. stdout: %s stderr: %s") % (
                                          cmd, proc.returncode, stdout, stderr))

            log.d("cmd '%s' stdout: %s, stderr: %s" % (cmd, stdout, stderr))
            return stdout.decode('utf-8')

    except err.ConfigCmdError:
        raise

    except Exception:
        raise err.ConfigCmdError("Error running command '%s': %s" % (
                                 cmd, traceback.format_exc())) from None

def get_rights(username, cellname, volid):
    # First, get the 1st-level membership for username
    pts = pts_auth()
    user_groups = pts.membership(nameorid=username).run()

    # Then, get the ACL for the root of the relevant volume
    fs = fs_auth()
    path = '/afs/.:mount/%s:%d' % (cellname, volid)
    acl = fs.listacl(path=path).run()

    cps = [username] + user_groups
    rights = {}

    # Now, go through all of the CPS for the user, and see what rights we gain
    for entry in acl.normal_rights:
        if entry.name in cps:
            for right in entry.rights:
                rights[right] = True

        elif entry.name in ('system:authuser', 'system:anyuser'):
            # s:authuser and s:anyuser will not show up via 'pts mem'. Let the
            # user inherit 'r' and 'l' rights here, in case they don't have
            # them otherwise, but do not let them inherit write-level access
            # from these.
            if 'r' in entry.rights:
                rights['r'] = True
            if 'l' in entry.rights:
                rights['l'] = True

    # Remove all rights applicable from negative acl entries.
    for entry in acl.negative_rights:
        if entry.name in cps or entry.name in ('system:authuser', 'system:anyuser'):
            for right in entry.rights:
                if right in rights:
                    del rights[right]

    return list(rights.keys())

def check_restore_authz(username, cellname, volid):
    log.d("Checking rights for user %s cell %s volid %d" % (
          username, cellname, volid))

    rights = get_rights(username, cellname, volid)

    log.d("User %s has rights %s on volid %d" % (username, rights, volid))

    for req_right in 'rlidw':
        if req_right not in rights:
            log.warn('no_access_info', "User %s does not have access right '%s' for volume id %d" % (
                                       username, req_right, volid))
            raise err.AccessError("Access denied for user %s" % username)

def volume_exists(volname, cell):
    vos = vos_unauth(cell=cell)
    cursor = vos.listvldb(name=volname)

    try:
        cursor.run()
    except fabs.cmd.ProcessError as exc:
        # If we got an error from listvldb, verify that the error that we got was
        # just that the volume didn't exist. Otherwise, raise the error we got.
        stderr = exc.stderr
        if stderr is not None and "VLDB: no such entry" in stderr:
            return False

        raise

    # No error from listvldb, so volume must exist
    return True

def path_exists(path):
    """
    Determine if `path` exists, without using stat()/lstat(). Instead we look
    at the directory listing for the parent dir, and see if an entry matches
    the given path. This is helpful for situations where stat'ing the actual
    path may hang, or error out with ENODEV, etc.
    """
    (head, tail) = os.path.split(path)
    if not head or not tail:
        raise err.InternalError("path_exists given '%s', which splits into (%s, %s)" % (
                                path, head, tail))

    entries = os.listdir(head)
    if tail in entries:
        return True

    return False

_alrm_set = False
@contextlib.contextmanager
def syscall_timeout(timeout):
    """
    Wrap a syscall in this to effectively get a timeout on the call. Example:

        with syscall_timeout(5):
            fcntl.flock(fh, fcntl.LOCK_EX)
            lock_success = True
        if not lock_success:
            timed_out()

    Since this is implemented with SIGALRM, you can only have one running at
    once, so you can't nest these. But the idea is just to wrap a single
    syscall, so the with: block for this should be very small anyway.
    """

    global _alrm_set
    if _alrm_set:
        raise err.InternalError("nested syscall_timeout call")

    class AlarmTriggered(Exception): pass
    def handler(signum, frame):
        raise AlarmTriggered

    orig_alrm = signal.signal(signal.SIGALRM, handler)

    try:
        _alrm_set = True
        signal.alarm(timeout)
        yield
    except AlarmTriggered:
        return
    finally:
        _alrm_set = False
        signal.alarm(0)
        signal.signal(signal.SIGALRM, orig_alrm)

def time_str2unix(timestr):
    """
    Parse a given datetime string, and return a unix timestamp.
    """
    if timestr[0] == '@':
        return int(timestr[1:])

    dt = dateutil.parser.parse(timestr)
    return int(dt.timestamp())

def time_unix2str(seconds):
    """
    Convert a unix timestamp into a human-readable datetime string.
    """
    return time.strftime("%a %b %d %H:%M:%S %Y %Z", time.localtime(seconds))

def chomp(msg, sep='\n'):
    """
    Trim off the trailing newline (or trailing sep') of 'msg', if it has one.
    """
    if msg[-1] == sep:
        return msg[:-1]
    return msg

def format_exc(exc):
    # Format the given exception into two strings, showing the backtrace
    # information, and the exception message.
    #
    # This is a bit cumbersome, because the traceback module does not give
    # convenient methods for formatting the full stack (along with chained
    # exceptions, etc) separately from the format_exception_only string. So, we
    # calculate the format_exception_only string, and just subtract those lines
    # from the full format_exception string to get the backtrace portion.
    msg_lines = list(traceback.format_exception_only(type(exc), exc))
    full_lines = list(traceback.format_exception(type(exc), exc,
                                                 exc.__traceback__))
    stack_lines = full_lines[:-len(msg_lines)]

    stack_str = chomp(''.join(stack_lines))
    msg_str = chomp(''.join(msg_lines))

    return (stack_str, msg_str)

class json_generator(list):
    """
    Wrap a generator in this 'list' subclass, in order to let json.dump()
    encode the generator like a normal list. An alternative to this is to use
    simplejson.dump(iterable_as_array=True), but using this lets us use the
    standard json module.

    This effectively just iterates over the given generator, but also appends
    something to the internal list if the generator is non-empty. This is
    important, because json needs __len__ to return nonzero in order to render
    the list at all, but __len__ must return 0 for an empty list.
    """

    # pylint: disable=super-init-not-called
    def __init__(self, gen):
        it = iter(gen)
        try:
            self._first = iter([next(it)])
            self.append(it)
        except StopIteration:
            self._first = []

    def __iter__(self):
        return itertools.chain(self._first, *self[:1])

def list2str(items):
    """
    Stringify a list like ['foo', 'bar', 'baz'] into a human-readable string
    like 'foo, bar, and baz'.
    """
    if len(items) > 1:
        items = items[:-1] + ['and ' + items[-1]]
    if len(items) > 2:
        return ', '.join(items)
    else:
        return ' '.join(items)
